/*
 * Copyright (C) 2017-2021 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

const { merge } = require("webpack-merge");
const fs = require("fs");
const HtmlWebpackPlugin = require("html-webpack-plugin");
const CopyPlugin = require("copy-webpack-plugin");
const copyfiles = require("copyfiles");
const webpack = require("webpack");
const MonacoWebpackPlugin = require("monaco-editor-webpack-plugin");

const HTML_ONLY = "html_only";

const DIST_FOLDER_PATH = process.env.DIST || "dist";
const VERSION = process.env.VERSION || "row";

function onCopyError(err) {
    if (err !== undefined) {
        console.log(err);
    }
}

const commonConfig = {
    mode: process.env.NODE_ENV || "development",
    context: __dirname,
    devtool: process.env.NODE_ENV !== "production" ? "source-map" : undefined,
    resolve: {
        extensions: [".webpack.js", ".web.js", ".web.ts", ".ts", ".tsx", ".js"],
        modules: [".", "coresdk", "node_modules"]
    },
    module: {
        rules: [
            { test: /\.tsx?$/, loader: "ts-loader" },
            { test: /\.(png|jpe?g|gif|svg)$/i, type: "asset/inline" },
            { test: /\.css$/, use: ["style-loader", "css-loader"] },
            { test: /\.s[ac]ss$/i, use: ["style-loader", "css-loader", "sass-loader"] },
            {
                test: /\.ttf$/,
                type: "asset",
                generator: {
                    filename: DIST_FOLDER_PATH + "/monaco-editor/" + "[name][ext]"
                }
            }
        ]
    },
    plugins: [
        new CopyPlugin({
            patterns: [{
                from: "src/*.js",
                to: DIST_FOLDER_PATH + "/" + VERSION + "/[name][ext]"
            }]
        })
    ],
    performance: {
        hints: false
    },
    output: {
        path: __dirname
    },
    optimization: {
        concatenateModules: false
    }
};

const examplesAppConfig = merge(commonConfig, {
    entry: {
        "examples": "./examples.ts"
    },
    output: {
        filename: DIST_FOLDER_PATH + "/[name].bundle.js"
    },
    devServer: {
        webSocketServer: 'sockjs',
        static: {
            directory: __dirname,
        }
    },
    externals: [
        {
            three: "THREE"
        },
        function ({context, request}, callback) {
            return /three\.module\.js$/.test(request) ? callback(null, "THREE") : callback();
        }
    ],
    resolve: {
        fallback: {
            stream: require.resolve("stream-browserify"),
            path: false
        }
    }
});

const decodersConfig = merge(commonConfig, {
    target: "webworker",
    entry: {
        "decoder": "./decoder/decoder.ts"
    },
    output: {
        filename: DIST_FOLDER_PATH + "/[name].bundle.js"
    },
    externals: [
        {
            three: "THREE"
        },
        function ({context, request}, callback) {
            return /three\.module\.js$/.test(request) ? callback(null, "THREE") : callback();
        }
    ],
    resolve: {
        fallback: {
            stream: require.resolve("stream-browserify"),
            path: false
        }
    }
});

const browserConfig = merge(commonConfig, {
    entry: {},
    output: {
        filename: DIST_FOLDER_PATH + "/" + VERSION + "/[name]_bundle.js"
    },
    externals: [
        {
            three: "THREE"
        },
        function ({context, request}, callback) {
            return /three\.module\.js$/.test(request) ? callback(null, "THREE") : callback();
        }
    ],
    resolve: {
        fallback: {
            stream: require.resolve("stream-browserify"),
            path: false
        }
    }
});

// moving common dependencies to the separate shared chunk
browserConfig.optimization = {
    ...browserConfig.optimization,
    splitChunks: {
        // include all types of chunks
        chunks: "all",
        name: "common_chunks"
    }
};

function getTemplate(chunk) {
    // exceptions
    const template = "template/" + chunk + ".html";
    if (fs.existsSync(template)) {
        return template;
    }

    // common templates
    if (chunk.indexOf("DI") > 0
        || chunk.indexOf("sensoris") > 0
        || chunk.indexOf("local_partition_development") > 0
        || chunk.indexOf("omv_plus_v2") > 0
    ) {
        return "template/example-DI.html";
    } else {
        return "template/example.html";
    }
}

function generateConfigs(examples, env) {
    for (let entry in examples) {
        browserConfig.entry[entry] = examples[entry].path;
    }

    browserConfig.plugins = Object.keys(browserConfig.entry).map(chunk =>
        new HtmlWebpackPlugin({
            template: getTemplate(chunk),
            chunks: ["common_chunks", chunk],
            filename: DIST_FOLDER_PATH + `/` + VERSION + `/${chunk}.html`
        })
    );

    browserConfig.plugins.push(
        new MonacoWebpackPlugin({
            // languages: ["typescript", "javascript"], // use only JS+TS languages
            filename: DIST_FOLDER_PATH + "/monaco-editor/[name].[ext].bundle.js"
        }),
        // Full set of moment.js locales is ~300 kB, and we use only Japanese and Chinese ones
        new webpack.ContextReplacementPlugin(/moment[/\\]locale$/, /ja|zh-cn/)
    );

    //IP: hack to bypass webpack 5 missed node polyfills
    browserConfig.plugins.push(
        new webpack.ProvidePlugin({
            Buffer: ['buffer', 'Buffer'],
        }));

    const files = { "highlighted": [] };

    const sectionHighlighted = [];
    for (let entry in browserConfig.entry) {
        let entryName = entry.split("_")[0];
        let section = files[entryName];
        if (section === undefined) {
            section = [];
        }

        const exampleProperties = {
            file: entry,
            devMode: examples[entry].devMode || false,
            viewSrc: examples[entry].viewSrc || null,
            isHTMLOnly: browserConfig.entry[entry] === HTML_ONLY,
            isNew: examples[entry].isNew || false,
        };
        if (examples[entry].title !== undefined) {
            exampleProperties.title = examples[entry].title;
            sectionHighlighted.push(exampleProperties);
        } else {
            section.push(exampleProperties);
            files[entryName] = section;
        }

        if (browserConfig.entry[entry] === HTML_ONLY) {
            delete browserConfig.entry[entry];
        }
    }

    if (sectionHighlighted.length) {
        files["highlighted"] = sectionHighlighted;
    }

    const licenseHeader = `/*
* Copyright (C) 2020-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
*
* This software and other materials contain proprietary information controlled by HERE and are
* protected by applicable copyright legislation. Any use and utilization of this software and other
* materials and disclosure to any third parties is conditional upon having a separate agreement
* with HERE for the access, use, utilization or disclosure of this software. In the absence of such
* agreement, the use of the software is not allowed.
*/

`;

    fs.writeFileSync(
        `lib/${env}/files.ts`,
        licenseHeader + `export const files = ${JSON.stringify(files, true, 4)};
`
    );

    copyfiles(["./resources/**/*", DIST_FOLDER_PATH + "/resources/"], { up: true }, onCopyError);

    return [decodersConfig, browserConfig, examplesAppConfig];
}

module.exports = { HTML_ONLY, generateConfigs };
