/*
 * Copyright (C) 2017-2021 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

/**
 * Renderer plugin to convert JSON data to GeoJSON.
 *
 * Generated GeoJSON must be created according to GeoJSON Specification:
 * https://tools.ietf.org/html/rfc7946
 *
 * This template is an example of plugin to generate the GeoJSON object based on decoded partition
 * content (the same data structure as in Decoded Data panel).
 *
 * In order to use renderer plugin this file must be placed into project root directory and
 * parameter `externalRendererPluginName` must be passed into constructor:
 *
 * ```
 * const geoJsonDataSource = new GeoJsonDataSource({
 *    dataStore: {
 *        hrn: HRN.fromString("Your HRN string"),
 *        getBearerToken: tokenResolver.getBearerToken.bind(tokenResolver),
 *        layer: "Layer name"
 *    },
 *    externalRendererPluginName: "renderer.plugin.template"
 * });
 * ```
 *
 * Renderer plugin must return an object with one of these functions in it:
 *
 * ```
 * getGeoJSON: function (params) { ... }
 * getGeoJSONAsync: async function (params) { ... }
 * ```
 *
 * Both synchronous \`getGeoJSON\` and async \`getGeoJSONAsync\` functions return an object with
 * GeoJSON data.
 *
 * Only one of these functions must be declared, if both are present plugin will not be used.
 */
(() => {
    "use strict";

    /** @type {RendererPlugin} */
    const plugin = {
        /**
         * Generates GeoJSON from RIB topology layer data.
         *
         * @param {object} params.logger Logger to log information to console.
         * @param {object} params.layer All properties as well as dataUrl, name, partitioning and
         * originalConfiguration.
         * @param {object} params.decodeResult Input RIB Navigation Attributes data.
         * @param {object} params.mapUtils Various converters like `mercatorConverter`,
         * `heretileToQuad`, `hdmapCoordinateToLonLat`, etc.
         * @param {object} params.geometryDataProviders Various geometry data providers to map data.
         * @returns {object} An object with generated GeoJSON data.
         */
        getGeoJSON: function (params) {
            // Destructuring of `params` object.
            const { logger, decodeResult } = params;

            const data = decodeResult.data;

            if (!data.segment) {
                throw new Error("Source data list of segments. Make sure you are passing correct data to the renderer.");
            }

            var result = {
                "type": "FeatureCollection",
                "features": data.segment.map(processLineStringArray)
            };
            logger.info("Data successfully processed: " + data.segment.length);

            // return object with GeoJSON
            return {
                contentType: "application/geo+json; charset=utf-8",
                body: result
            };
        }
    };

    return plugin;

    function processLineStringArray(segment, index) {
        const coordinates = segment.geometry.point.map(point => [point.longitude, point.latitude]);
        return {
            type: "Feature",
            properties: {
                protobufRef: `segment[${index}]`,
                identifier: segment.identifier,
                length: segment.length,
                width: 5
            },
            geometry: {
                type: "LineString",
                coordinates
            }
        };
    }
})();
