/*
 * Copyright (C) 2020-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

/** @module @here/olp-examples
 * [[include:olp-examples/README.md]]
 *  **//** */

//IP: for some reason import is not working for highlight.js 11.1
const hljs = require("highlight.js/lib/core");
const typescript = require("highlight.js/lib/languages/typescript");
hljs.registerLanguage('typescript', typescript);
import "highlight.js/styles/default.css";
import { HighlightResult } from "highlight.js";

const nunjucks = require("nunjucks");
nunjucks.configure({ autoescape: true });

import { files as filesCN } from "../lib/cn/files";
import { files as filesROW } from "../lib/row/files";
import { I18n } from "../lib/i18n/I18n";

import "../lib/di-examples.scss";

/**
 * An interface for a particular example file.
 */
export interface ExampleFile {
    /** File name without extension. */
    file: string;
    /** Indicates that example has to be visible only in dev mode. */
    devMode: boolean;
    /** Allows to set specific source file path instead of auto generated. */
    viewSrc: string | null;
    /** Indicates that there is no TypeScript source file for given example. */
    isHTMLOnly: boolean;
    /** Allows to set specific title in examples list. */
    title?: string;
    /** Indicates that example is new and requires a label `New` to be shown. */
    isNew: boolean;
}

/**
 * An interface for a list of examples grouped by sections.
 */
export interface FilesList {
    /** A group of examples. */
    [propName: string]: ExampleFile[];
}

/**
 * A version of examples adapted to specific region.
 */
export enum DIExamplesEnvironment {
    ROW = "Rest of the World",
    CN = "China"
}

/**
 * Properties of Data Inspector Examples engine.
 */
export interface DIExamplesProps {
    /** Path to a folder where examples application located. Useful when examples are part of
     *  another application.
     *  @default ""
     */
    examplesPath?: string;
    /**
     * Path to a folder where compiled examples are located. Relative to examplesPath.
     * @default "dist/"
     */
    distPath?: string;
    /**
     * Environment.
     * @default DIExamplesEnvironment.ROW
     */
    env?: DIExamplesEnvironment;
    /**
     * Path to a folder where source files exist. Relative to examplesPath.
     * @default "src/"
     */
    sourcesPath?: string;
    /**
     * Path to a folder where files with descriptions of examples exist. Relative to examplesPath.
     * @default "docs/"
     */
    docsPath?: string;
    /**
     * An URL to examples archive.
     * @default "https://platform.here.com/data-inspector/data-inspector-examples-latest.zip"
     */
    downloadExamplesURL?: string;
    /**
     * Environments of the Lookup Service.
     */
    lookupEnvironment?: string;
    /**
     * User access token.
     */
    hereUserAccessToken?: string;
    /**
     * UI locale with ISO 639-1 language code and ISO-3166 country code pair.
     * @default "en-US"
     */
    locale?: string;
    /**
     * A path to Izumi translation JSON files.
     * @default "dist/resources/i18n/"
     */
    translationsBasePath?: string;
    /**
     * Show or hide the language selector widget.
     */
    showLanguageSelector?: boolean;
    /**
     * false to have possibility to open specific example in new tab without `DIExamples`
     * wrapper. Useful when running examples locally. Set to true to be able to still be logged
     * in when open such example in new tab on HERE platform.
     * @default false
     */
    openWithWrapper?: boolean;
}

/**
 * `DIExamples` is an engine of Data Inspector examples which builds application with examples list,
 * live examples and additional details of each example. This class allows to build a standalone
 * application with examples as well as easily integrate examples into another complex solution.
 */
export class DIExamples {
    /** A collection of links generated for all examples. */
    private m_links: Map<string, HTMLAnchorElement> = new Map();
    /** Selected example file name. */
    private m_selected: string | null = null;
    /** An array of all examples not grouped into different sections. */
    private m_allFiles: ExampleFile[] = [];
    /** IFrame for live example presentation. */
    private m_viewer?: HTMLIFrameElement;
    /** View source code button. */
    private m_viewSourceBtn?: HTMLAnchorElement;
    /** Example details button. */
    private m_openDetailsBtn?: HTMLAnchorElement;
    /** Options needed to replace template keys to values in example details files. */
    private m_gitBookOptions?: object;
    /** Example name title. */
    private m_exampleTitle?: HTMLDivElement;
    /** Button which opens examples list. */
    private m_openListBtn?: HTMLAnchorElement;
    /** Left panel container. */
    private m_leftPanel?: HTMLDivElement;
    /** Example details panel container. */
    private m_detailsPanel?: HTMLDivElement;
    /** Examples list container. */
    private m_examplesList?: HTMLDivElement;
    /** Path to a folder where compiled examples application located. */
    private m_examplesPath: string;
    /** Path to a folder where compiled examples are located. Relative to m_examplesPath. */
    private m_distPath: string;
    /** Environment. */
    private m_env: DIExamplesEnvironment;
    /** Path to a folder where source files exist. Relative to m_examplesPath. */
    private m_sourcesPath: string;
    /** Path to a folder where files with descriptions of examples exist. Relative to m_examplesPath. */
    private m_docsPath: string;
    /** An URL to examples archive. */
    private m_downloadExamplesURL: string;
    /** Additional data passed to example in iframe through GET params. */
    private m_additionalGetParams?: string;
    /** UI locale with ISO 639-1 language code and ISO-3166 country code pair. */
    private m_locale: string;
    /** A path to Izumi translation JSON files. */
    private m_translationsBasePath: string;
    /** User access token. */
    private m_hereUserAccessToken?: string;
    /** Handles providing of user token to iframe. */
    private m_postMessageHandler?: (event: MessageEvent) => void;
    /**
     * Defines whether specific example is opened in new tab with or without `DIExamples`
     * wrapper.
     */
    private m_openWithWrapper: boolean;


    /**
     * A default constructor.
     *
     * @param m_container Container where examples has to be placed.
     * @param props Properties of examples application.
     */
    constructor(
        private m_container: HTMLElement,
        props: DIExamplesProps = {}
    ) {
        // set default values
        this.m_examplesPath = props.examplesPath || "";
        this.m_distPath = this.m_examplesPath + (props.distPath || "dist/");
        this.m_env = props.env || DIExamplesEnvironment.ROW;
        this.m_sourcesPath = this.m_examplesPath + (props.sourcesPath || "src/");
        this.m_docsPath = this.m_examplesPath + (props.docsPath || "docs/");
        this.m_translationsBasePath = this.m_examplesPath + (props.translationsBasePath || "dist/resources/i18n/");
        this.m_locale = props.locale || I18n.DEFAULT_LOCALE;
        this.m_downloadExamplesURL = props.downloadExamplesURL
            || "https://platform.here.com/data-inspector/data-inspector-examples-latest.zip";
        this.m_hereUserAccessToken = props.hereUserAccessToken;
        this.m_openWithWrapper = props.openWithWrapper || false;

        const params = [];
        if (props.lookupEnvironment !== undefined) {
            params.push("lookupEnvironment=" + props.lookupEnvironment);
        }
        if (props.hereUserAccessToken !== undefined) {
            params.push("useToken=true");
        }
        if (this.m_openWithWrapper === true) {
            params.push("hideURL=true");
        }
        if (props.locale !== undefined) {
            params.push("locale=" + props.locale);
        }
        if (props.showLanguageSelector !== undefined) {
            params.push("showLanguageSelector=" + props.showLanguageSelector);
        }
        this.m_additionalGetParams = "?" + params.join("&");

        // GET parameter `devMode` provides ability to see hidden examples created for development
        // purpose.
        if (new URL(window.location.href).origin.indexOf("olp.pages.gitlab.in.here.com") !== -1) {
            window.localStorage.setItem("olp-examples-dev-mode", "true");
        }
        const devMode = new URL(window.location.href).searchParams.get("devMode");

        if (devMode && ["true", "false"].includes(devMode)) {
            window.localStorage.setItem("olp-examples-dev-mode", devMode);
        }

        this.mountApp();
    }

    /**
     * Cleans the resources created for DIExamples.
     */
    destroy(): void {
        if (this.m_postMessageHandler !== undefined) {
            window.removeEventListener("message", this.m_postMessageHandler);
        }
    }

    /**
     * Initializes internationalization and generates UI.
     */
    private async mountApp(): Promise<void> {
        await this.initializeI18n();

        // show examples or source code
        const src = this.getSourceFromLocation();
        if (src === null || src.length === 0) {
            this.createPageLayout();
            this.addExamplesList();
        } else {
            this.showSource(src);
        }
    }

    /**
     * Initializes internationalization.
     */
    private async initializeI18n(): Promise<void> {
        const locale = this.m_locale;
        if (locale !== I18n.instance.locale) {
            try {
                await I18n.loadTranslation(locale, this.m_translationsBasePath);
            } catch {
                console.warn(`Cannot retrieve JSON file for translation ${locale}`);
            }
        }
    }

    /**
     * Generates and renders page layout.
     */
    private createPageLayout(): void {
        // General examples layout
        const examplesWrapper = document.createElement("div");
        examplesWrapper.classList.add("examples-wrapper");
        this.m_container.appendChild(examplesWrapper);

        this.m_leftPanel = document.createElement("div");
        this.m_leftPanel.classList.add("left-panel");
        examplesWrapper.appendChild(this.m_leftPanel);

        const mainContainer = document.createElement("div");
        mainContainer.classList.add("main-container");
        examplesWrapper.appendChild(mainContainer);


        // Left Panel

        // Header container
        const leftPanelHeader = document.createElement("div");
        leftPanelHeader.classList.add("header");
        this.m_leftPanel.appendChild(leftPanelHeader);

        // Examples title
        const listTitle = document.createElement("h1");
        listTitle.textContent = I18n.translate("olpdi.examples.engine-app.menu.title");
        leftPanelHeader.appendChild(listTitle);

        // Download examples button
        const downloadBtn = document.createElement("a");
        downloadBtn.classList.add("download-btn");
        downloadBtn.title = I18n.translate("olpdi.examples.engine-app.download-examples-tooltip");
        downloadBtn.href = this.m_downloadExamplesURL;
        leftPanelHeader.appendChild(downloadBtn);

        // Empty space
        const space = document.createElement("div");
        space.classList.add("space");
        leftPanelHeader.appendChild(space);

        // Close panel button
        const closeLeftPanelBtn = document.createElement("a");
        closeLeftPanelBtn.classList.add("close-left-panel-btn");
        closeLeftPanelBtn.title = I18n.translate("olpdi.examples.engine-app.menu.close-tooltip");
        closeLeftPanelBtn.onclick = () => this.toggleMenu();
        leftPanelHeader.appendChild(closeLeftPanelBtn);

        // Examples list container
        this.m_examplesList = document.createElement("div");
        this.m_examplesList.classList.add("list");
        this.m_leftPanel.appendChild(this.m_examplesList);


        // Main Container

        // Header
        const mainContainerHeader = document.createElement("div");
        mainContainerHeader.classList.add("header");
        mainContainer.appendChild(mainContainerHeader);

        // Open Left Panel button
        this.m_openListBtn = document.createElement("a");
        this.m_openListBtn.classList.add("open-list-btn");
        this.m_openListBtn.title = I18n.translate("olpdi.examples.engine-app.menu.open-tooltip");
        this.m_openListBtn.onclick = () => this.toggleMenu();
        mainContainerHeader.appendChild(this.m_openListBtn);

        // Example name title
        this.m_exampleTitle = document.createElement("div");
        this.m_exampleTitle.classList.add("example-title");
        mainContainerHeader.appendChild(this.m_exampleTitle);

        // View source button
        this.m_viewSourceBtn = document.createElement("a");
        this.m_viewSourceBtn.textContent = I18n.translate("olpdi.examples.engine-app.open-source-code");
        this.m_viewSourceBtn.classList.add("view-source-btn");
        this.m_viewSourceBtn.onclick = () => {
            if (this.m_selected === null) {
                return;
            }
            let scriptFile: string | null = null;
            if (this.m_links.has(this.m_selected)) {
                const link = this.m_links.get(this.m_selected) as HTMLAnchorElement;
                scriptFile = link.getAttribute("data-src");
            }
            if (!scriptFile) {
                scriptFile = this.m_selected + ".ts";
            }
            window.open(location.pathname + "?src=" + encodeURIComponent(scriptFile));
        };
        mainContainerHeader.appendChild(this.m_viewSourceBtn);

        // Example details
        const exampleDetailsContainer = document.createElement("div");
        exampleDetailsContainer.classList.add("example-details");
        mainContainerHeader.appendChild(exampleDetailsContainer);

        // Example details button
        this.m_openDetailsBtn = document.createElement("a");
        this.m_openDetailsBtn.textContent = I18n.translate("olpdi.examples.engine-app.example-details");
        this.m_openDetailsBtn.classList.add("open-details-btn");
        this.m_openDetailsBtn.onclick = () => this.toggleDetails();
        exampleDetailsContainer.appendChild(this.m_openDetailsBtn);

        // Example details panel
        this.m_detailsPanel = document.createElement("div");
        this.m_detailsPanel.classList.add("details-panel");
        exampleDetailsContainer.appendChild(this.m_detailsPanel);

        // Example viewer iframe
        this.m_viewer = document.createElement("iframe");
        this.m_viewer.classList.add("viewer");
        this.m_viewer.name = "viewer";
        this.m_viewer.allowFullscreen = true;
        this.m_viewer.setAttribute("allowvr", "");
        this.m_viewer.setAttribute("onmousewheel", "");
        mainContainer.appendChild(this.m_viewer);

        this.m_postMessageHandler = (event: MessageEvent) => {
            const eventSource = event.source as Window;
            if (event.data.type !== "tokenRequest"
                || event.source === null
                || eventSource.name !== "viewer"
                || this.m_viewer === undefined
                || this.m_viewer.contentWindow === null) {
                return;
            }

            this.m_viewer.contentWindow.postMessage({
                type: "userToken",
                token: this.m_hereUserAccessToken
            }, "*");
        };

        window.addEventListener("message", this.m_postMessageHandler);
    }

    /**
     * Creates and adds the list of examples to the page.
     */
    private addExamplesList(): void {
        if (this.m_examplesList === undefined) {
            return;
        }

        // use proper version of examples
        const files = (this.m_env === DIExamplesEnvironment.CN) ? filesCN : filesROW;

        for (const key in files) {
            const section = (files as FilesList)[key];
            const sectionLength = section.filter(f => !f.devMode || this.isDevMode()).length;
            if (sectionLength === 0) {
                continue;
            }
            this.m_allFiles = this.m_allFiles.concat(section);

            const sectionName = document.createElement("h2");
            sectionName.textContent = I18n.translate("olpdi.examples.engine-app.section-" + key);
            this.m_examplesList.appendChild(sectionName);

            const count = document.createElement("span");
            count.classList.add("count");
            count.textContent = ` (${sectionLength})`;
            sectionName.appendChild(count);

            const ul = document.createElement("ul");
            this.m_examplesList.appendChild(ul);
            section.forEach(file => {
                if (file.devMode && !this.isDevMode()) {
                    return;
                }

                const li = document.createElement("li");
                ul.appendChild(li);
                const link = this.createLink(file);
                li.appendChild(link);

                this.m_links.set(file.file, link);
            });
        }

        this.checkURL();
    }

    /**
     * Indicates whether development mode is active or not.
     */
    private isDevMode(): boolean {
        return window.localStorage.getItem("olp-examples-dev-mode") === "true";
    }

    /**
     * Toggles panel with examples list.
     */
    private toggleMenu(): void {
        if (this.m_openListBtn === undefined || this.m_leftPanel === undefined) {
            return;
        }
        this.m_openListBtn.style.display = (this.m_leftPanel.style.display !== "none")
            ? "inherit"
            : "none";
        this.m_leftPanel.style.display = (this.m_leftPanel.style.display !== "none")
            ? "none"
            : "inherit";
    }

    /**
     * Toggles panel with example details.
     */
    private toggleDetails(): void {
        if (this.m_detailsPanel === undefined) {
            return;
        }

        if (this.m_detailsPanel.style.display === "block") {
            this.closeDetailsPanel();
        } else {
            this.openDetailsPanel();
        }
    }

    /**
     * Opens details panel.
     */
    private openDetailsPanel(): void {
        if (this.m_openDetailsBtn === undefined || this.m_detailsPanel === undefined) {
            return;
        }

        this.m_openDetailsBtn.classList.add("active");
        this.m_detailsPanel.style.display = "block";
    }

    /**
     * Closes details panel.
     */
    private closeDetailsPanel(): void {
        if (this.m_openDetailsBtn === undefined || this.m_detailsPanel === undefined) {
            return;
        }

        this.m_openDetailsBtn.classList.remove("active");
        this.m_detailsPanel.style.display = "none";
    }

    /**
     * Shows details button.
     */
    private showDetailsBtn(): void {
        if (this.m_openDetailsBtn === undefined) {
            return;
        }

        this.m_openDetailsBtn.style.display = "block";
    }

    /**
     * Hides details button.
     */
    private hideDetailsBtn(): void {
        if (this.m_openDetailsBtn === undefined) {
            return;
        }

        this.m_openDetailsBtn.style.display = "none";
    }

    /**
     * Reads example details from file and puts into details panel.
     */
    private async getExampleDetails(): Promise<void> {
        if (this.m_detailsPanel === undefined) {
            return;
        }

        const detailsFile = this.m_docsPath + this.m_selected + ".md";
        const response = await fetch(detailsFile, { credentials: "same-origin" });
        let details = "";
        if (response.ok) {
            // read details
            details = await response.text();

            // replace template keys with variable values
            const gitBookOptions = await this.getGitBookOptions() as any;
            const filterValue = (this.m_env === DIExamplesEnvironment.ROW) ? "olp" : "olp-chn";
            const filtered = gitBookOptions.variables.filtered
                .find((object: any) => object.filter === filterValue);

            details = nunjucks.renderString(details, {
                "book": {
                    "standard": {
                        ...gitBookOptions.variables.standard
                    },
                    "filtered": {
                        ...filtered.variables
                    }
                }
            });

            // show details button
            this.showDetailsBtn();
        } else {
            // hide details panel and button
            this.hideDetailsBtn();
            this.closeDetailsPanel();
        }

        this.m_detailsPanel.innerHTML = details;
    }

    /**
     * Fetches content of book.json file with template keys and values which have to be
     * replaced in examples details.
     */
    private async getGitBookOptions(): Promise<object> {
        if (this.m_gitBookOptions !== undefined) {
            return this.m_gitBookOptions;
        }

        const optionsFile = this.m_docsPath + "book.json";
        const response = await fetch(optionsFile, { credentials: "same-origin" });
        if (!response.ok) {
            throw new Error(`Failed retrieving file "${optionsFile}": ${response.statusText}`);
        }

        const options = await response.text();
        if (options === undefined) {
            throw new Error(`Gitbook configuration file is empty`);
        }
        this.m_gitBookOptions = JSON.parse(options) as object;

        return this.m_gitBookOptions;
    }

    /**
     * Loads source code from example file and renders it with highlighting.
     *
     * @param src Path to example file.
     */
    private async showSource(src: string) {
        const codeElement = document.createElement("pre");
        codeElement.classList.add("code");
        this.m_container.appendChild(codeElement);

        if (this.m_env === DIExamplesEnvironment.CN) {
            src = src.replace("../config.ts", "../config.cn.ts");
        }

        const baseDir = src.substring(0, src.lastIndexOf("/"));

        src = (src.includes("dist/"))
            // replace default "dist/" to "dist-cn/" if one was changed
            ? src.replace("dist/", this.m_distPath)
            // set path to sources folder
            : this.m_sourcesPath + src;
        const response = await fetch(src, { credentials: "same-origin" });
        if (!response.ok) {
            throw new Error(`Failed retrieving file "${src}": ${response.statusText}`);
        }

        const contents = await response.text();

        const highlightedContents: HighlightResult = hljs.highlight("typescript", contents, true);
        const importRe = /(from.*(?:'|"))(\..*?)('|")/g;
        codeElement.innerHTML = highlightedContents.value.replace(importRe, (_, p1, p2, p3) => {
            let path = p2 + ".ts";
            if (baseDir !== "") {
                path = (baseDir + "/" + path).replace("/./", "/");
            }
            // tslint:disable-next-line:max-line-length
            return `${p1}<a href="${location.pathname}?src=${encodeURIComponent(path)}">${p2}</a>${p3}`;
        });
    }

    /**
     * Checks whether webpage address contains parameter with path to example source file and
     * returns it if found.
     */
    private getSourceFromLocation(): string | null {
        if (location.search === undefined || location.search.length === 0) {
            return null;
        }

        const params = new URLSearchParams(location.search.slice(1));
        return params.get("src");
    }

    /**
     * Creates examples links for examples list section.
     *
     * @param file Example file details.
     */
    private createLink(file: ExampleFile): HTMLAnchorElement {
        const title = file.title || this.getName(file.file);
        const viewSrc = file.viewSrc || null;

        const link = document.createElement("a");

        link.href = this.getExampleUrl(file.file, this.m_openWithWrapper);
        link.textContent = title;
        link.setAttribute("target", "viewer");
        if (viewSrc) {
            link.setAttribute("data-src", viewSrc);
        }
        link.onclick = (event) => {
            if (event.button === 0) {
                // we need to prevent default behaviour because links are not always links to
                // direct examples https://<host>/dist/row/<example>.html but could be link to
                // example via DIExample app: https://<host>/#example
                event.preventDefault();
                this.selectFile(file);
            }
        };

        if (file.isNew) {
            const newLable = document.createElement("span");
            newLable.classList.add("new");
            newLable.textContent = I18n.translate("olpdi.examples.engine-app.new-example");
            link.appendChild(newLable);
        }

        return link;
    }

    /**
     * Checks whether webpage address contains name of example and opens such example.
     */
    private checkURL(): void {
        if (window.location.hash === "") {
            return;
        }

        const filename = window.location.hash.substring(1);
        const matchedFilenames = /[/\w]+/g.exec(filename);
        if (matchedFilenames === null) {
            return;
        }
        const file = this.m_allFiles.find(item => item.file === matchedFilenames[0]);
        if (file === undefined) {
            return;
        }
        this.selectFile(file);
    }

    /**
     * Prepares URL of example.
     * If needed adds additional GET parameters required for integration into Portal.
     *
     * @param filename Example file name.
     * @param keepWrapper defines whether keep wrapper when opening example in new tab.
     */
    private getExampleUrl(filename: string, keepWrapper: boolean): string {
        let url = keepWrapper ? this.m_examplesPath + "#" + filename : this.m_distPath + filename + ".html";
        if (this.m_additionalGetParams !== undefined) {
            url += this.m_additionalGetParams;
        }

        return url;
    }

    /**
     * Handles selection of example.
     *
     * @param file Example file details.
     */
    private selectFile(file: ExampleFile): void {
        // reset previously selected link
        if (this.m_selected !== null && this.m_links.has(this.m_selected)) {
            const link = this.m_links.get(this.m_selected) as HTMLAnchorElement;
            link.classList.remove("selected");

            this.m_links.set(this.m_selected, link);
        }

        const fileName = file.file;
        this.m_selected = fileName;

        // set new link as selected
        if (this.m_links.has(fileName)) {
            const link = this.m_links.get(fileName) as HTMLAnchorElement;
            link.classList.add("selected");

            this.m_links.set(fileName, link);
        }

        // change URL
        window.location.hash = fileName;

        if (this.m_viewer !== undefined) {
            this.m_viewer.focus();
        }

        // hide "Source code" link if there is nothing to show
        if (this.m_viewSourceBtn !== undefined) {
            this.m_viewSourceBtn.style.display = file.viewSrc === null && file.isHTMLOnly
                ? "none"
                : "block";
        }

        // show example name in header
        if (this.m_exampleTitle !== undefined) {
            this.m_exampleTitle.textContent = file.title || this.getName(file.file);
        }

        // update details panel with details of selected example
        this.getExampleDetails();

        // we change an iframe src to the direct example link
        if (this.m_viewer !== undefined) {
            this.m_viewer.src = this.getExampleUrl(file.file, false);
        }
    }

    /**
     * Returns example name generated from file name.
     *
     * @param file Example file name.
     */
    private getName(file: string): string {
        const name = file.split("_");
        name.shift();
        return name.join(" / ");
    }
}
