/*
 * Copyright (C) 2019-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

/** @module @here/olp-examples
 * [[include:olp-examples/README.md]]
 *  **//** */

import { DataStoreDownloadManager } from "@here/olp-sdk-core";
import i18next, { i18n } from "i18next/index.js";
import * as moment from "moment";
import { default as DefaultLocale } from "@here/interactive-mapview-ui/lib/i18n/DefaultLocale";

/**
 * Izumi localization JSON.
 */
export interface IzumiL10n {
    /** Options */
    options?: {
        /** Specifies whether language is RTL or not. */
        rtl: boolean;
    };

    /** List of translations. */
    data: {
        [key: string]: string;
    };
}

/**
 * I18n class provides internationalization and localization of string resources.
 */
export class I18n {
    /** A default locale, which is US English. */
    static readonly DEFAULT_LOCALE = "en-US";

    /**
     * Adds language.
     *
     * @param language Language identifier.
     * @param translations Translations.
     */
    static addLanguage(language: string, translations: IzumiL10n) {
        const instance = I18n.instance;
        for (const key in translations.data) {
            if (translations.data.hasOwnProperty(key)) {
                instance.addTranslation(language, key, translations.data[key]);
            }
        }
    }

    /**
     * Removes translations for certain language.
     *
     * @param language Language to remove.
     */
    static removeLanguage(language: string) {
        I18n.instance.removeTranslations(language);
    }

    /**
     * A shortcut method for non static method `translate`.
     *
     * @param key A translation key.
     * @param interpolationValues Interpolation values to be applied to the translated string.
     * @returns The translated string, or an empty string if translation isn't available for this
     * data source.
     */
    static translate(key: string, interpolationValues?: { [key: string]: any }): string {
        return I18n.instance.translate(key, interpolationValues);
    }

    /**
     * Asynchronously loads Izumi JSON file and sets this locale.
     *
     * @param locale Locale.
     * @param translationsBasePath Base path to translation JSON files.
     * @param removeCurrentLocale `true` to clear current locale after loading new one, `false` to
     * keep existing locale.
     */
    static async loadTranslation(
        locale: string,
        translationsBasePath: string,
        removeCurrentLocale = true
    ): Promise<void> {
        const previousLocale = I18n.instance.locale;

        if (locale === previousLocale) {
            return;
        }

        if (locale === I18n.DEFAULT_LOCALE) {
            I18n.addLanguage(I18n.DEFAULT_LOCALE, DefaultLocale);
        } else {
            const basePath = `${translationsBasePath}${translationsBasePath.endsWith("/") ? "" : "/"}`;
            const downloadManager = new DataStoreDownloadManager();
            const response = await downloadManager.download(`${basePath}${locale}.json`);
            const translations = await response.json() as IzumiL10n;
            I18n.addLanguage(locale, translations);
        }
        I18n.instance.locale = locale;

        if (removeCurrentLocale) {
            // Izumi translations for other languages are already merged with default values for
            // missing translations so we can clean up default translations.
            I18n.removeLanguage(previousLocale);
        }
    }

    /** Translation namespace for i18next. */
    private static readonly TRANSLATION_NS = "translation";

    /** Singleton instance. */
    private static m_instance: I18n | undefined;

    /** An instance of `i18next`. */
    private m_i18next: i18n | undefined;

    /** Instance of I18n. */
    static get instance(): I18n {
        if (I18n.m_instance === undefined) {
            I18n.m_instance = new I18n();
            I18n.addLanguage(I18n.DEFAULT_LOCALE, DefaultLocale);
            I18n.instance.locale = I18n.DEFAULT_LOCALE;
        }
        return I18n.m_instance;
    }

    /** Gets the current locale. */
    get locale(): string {
        return this.i18n.language;
    }

    /**
     * Changes the current locale.
     */
    set locale(value: string) {
        this.i18n.changeLanguage(value);
        moment.locale(value);
    }

    /** Constructor is private for this class to be used as a singleton. */
    private constructor() {}

    /**
     * Returns translation for a certain `key`. If translation isn't available in specific language,
     * the default language [[InteractiveDataSource.DEFAULT_LOCALE]] is used.
     *
     * @param key A translation key.
     * @param interpolationValues Interpolation values to be applied to the translated string.
     * @returns The translated string, or an empty string if translation isn't available for this
     * data source.
     * @example **Localization string with interpolation (added using [[addTranslation]]):**
     * ```typescript
     * "Hello, {{user.name}}"
     * ```
     *
     * **`interpolationValues` parameter:**
     * ```typescript
     * { user: { name: "John" } }
     * ```
     */
    translate(key: string, interpolationValues?: { [key: string]: any }): string {
        return this.i18n.t(key, interpolationValues);
    }

    /**
     * Adds a new translation value or modifies the existing one for the same key. For the
     * description of the interpolation format, see the [i18next
     * documentation](https://www.i18next.com/interpolation.html).
     *
     * @param lng A language for translation.
     * @param key A key.
     * @param value A value.
     */
    addTranslation(lng: string, key: string, value: string): void {
        this.i18n.addResource(lng, I18n.TRANSLATION_NS, key, value);
    }

    /**
     * Removes translations for certain language.
     *
     * @param language Language to remove.
     */
    removeTranslations(language: string) {
        this.i18n.removeResourceBundle(language, I18n.TRANSLATION_NS);
    }

    /** Initializes localization. */
    private get i18n(): i18n {
        if (this.m_i18next && this.m_i18next.isInitialized === true) {
            return this.m_i18next;
        }

        this.m_i18next = i18next.createInstance();

        this.m_i18next.init(
            {
                keySeparator: false,
                initImmediate: false,
                ns: I18n.TRANSLATION_NS,
                defaultNS: I18n.TRANSLATION_NS,
                fallbackLng: I18n.DEFAULT_LOCALE,
                debug: false,
                resources: {},
                saveMissing: true,
                missingKeyHandler: (lngs: readonly string[], ns: string, key: string, fallback: string) => {
                    console.warn(`No translation for key "${key}"`);
                }
            }
        );

        return this.m_i18next;
    }
}
