/*
 * Copyright (C) 2017-2021 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

/**
 * Renderer plugin to convert HMC content JSON data to GeoJSON.
 * Supports visualization of Road Topology and Geometry, Buildings and Cartography data.
 */
(() => {
    "use strict";

    /** Parameters for plugin. */
    let pluginParams;
    /** Promises for metadata providers. */
    let metadataProviderPromises = {};
    /** Cached metadata partitions. */
    let metadataCache = new Map();

    /** @type {RendererPlugin} */
    const plugin = {
        /**
         * Asynchronously generates GeoJSON from RIB Navigation Attributes data.
         *
         * @param {object} params.logger Logger to log information to console.
         * @param {object} params.layer All properties as well as dataUrl, name, partitioning and
         * originalConfiguration.
         * @param {object} params.decodeResult Input RIB Navigation Attributes data.
         * @param {object} params.mapUtils Various converters like `mercatorConverter`,
         * `heretileToQuad`, `hdmapCoordinateToLonLat`, etc.
         * @param {object} params.geometryDataProviders Various geometry data providers to map data.
         * @returns {Promise} A promise which resolves with generated GeoJSON data.
         */
        async getGeoJSONAsync(params) {
            pluginParams = params;

            const logger = params.logger;
            logger.level = 0;
            logger.info(`Creating GeoJSON for tile ${params.decodeResult.partitionId}`);
            const data = params.decodeResult.data;

            let features = [];

            if (data.node || data.segment) {
                features = features.concat(processSegments(data.segment || []));
                features = features.concat(processNodes(data.node || []));
            } else if (data.location && data.place) {
                let placesRootTag = "Places";
                if (params.layer.id === "building-footprints") {
                    placesRootTag = "Building Footprints";
                } else if (params.layer.id === "cartography") {
                    placesRootTag = "Cartography";
                }
                features = features.concat(await processPlaces(data, placesRootTag));
            } else {
                logger.error("Source data does not have HMC data. Make sure you are passing correct data to the renderer.");
            }

            return {
                contentType: "application/geo+json; charset=utf-8",
                body: {
                    type: "FeatureCollection",
                    features
                }
            };
        }
    };

    return plugin;

    /**
     * Returns a promise with metadata provider.
     *
     * @param {string} layer Metadata layer.
     */
    function getMetadataProvider(layer) {
        if (metadataProviderPromises[layer] === undefined) {
            metadataProviderPromises[layer] = pluginParams.getPartitionDataProvider(layer);
        }
        return metadataProviderPromises[layer];
    }

    /**
     * Returns a promise with metadata.
     *
     * @param {string} layer Metadata layer.
     * @param {string} partition Metadata partition id.
     */
    async function getMetadata(layer, partition) {
        const key = `${layer}|${partition}`;
        if (metadataCache.has(key)) {
            return metadataCache.get(key);
        }

        const provider = await getMetadataProvider(layer);
        const data = await provider.getDataDecoded(partition);
        metadataCache.set(key, data);
        return data;
    }

    /**
     * Returns a promise with a category.
     *
     * @param {*} id Category id.
     * @param {*} partition Metadata partition id.
     */
    async function getCategory(id, partition) {
        let metadataLayer;
        if (id.includes("here:cm:carto-category")) {
            metadataLayer = "cartography-metadata";
        } else if (id.includes("here:cm:building-category")) {
            metadataLayer = "building-metadata";
        } else {
            return undefined;
        }

        const metadata = await getMetadata(metadataLayer, partition);
        return metadata.category.find(item => item.identifier === id);
    }

    /**
     * Returns a promise with places geometries.
     *
     * @param {object} data Decoded places partition data.
     * @param {string} placesRootTag Root tag for features.
     */
    async function processPlaces(data, placesRootTag) {
        const features = [];
        const locations = new Map(data.location.map(item => [item.identifier, item]));

        for (let index = 0; index < data.place.length; index++) {
            const place = data.place[index];
            const location = locations.get(place.location_ref.identifier);
            if (location === undefined) {
                this.m_logger.warn(`Geometry for location "${place.location_ref.identifier}" not found`);
                return;
            }

            const geometry = location.geometry;
            if (
                geometry.multi_polygon === undefined
                && geometry.multi_line_string === undefined
            ) {
                this.m_logger.warn(`Unsupported geometry type`, geometry);
                return;
            }

            const placeCategories = await getPlaceCategories(place);
            const placeTags = await getPlaceCategories(place, placesRootTag);
            const name = getPlaceNames(place).join(", ");
            const properties = {
                description: {
                    identifier: place.identifier,
                    name,
                    category: placeCategories.join(",\n")
                },
                title: `${name !== "" ? name + "<br>" : ""}${place.identifier}`,
                protobufRef: `place[${index}]`,
                featureTags: placeTags
            };

            if (geometry.multi_polygon !== undefined) {
                features.push({
                    type: "Feature",
                    properties,
                    geometry: {
                        type: "MultiPolygon",
                        coordinates: geometry.multi_polygon.polygon.map(polygon => {
                            const result = [
                                polygon.exterior_ring.point.map(pointToGeoJsonCoordinate)
                            ];
                            polygon.interior_ring.forEach(ring =>
                                result.push(ring.point.map(pointToGeoJsonCoordinate))
                            );
                            return result;
                        })
                    }
                });
            }

            if (geometry.multi_line_string !== undefined) {
                features.push({
                    type: "Feature",
                    properties,
                    geometry: {
                        type: "MultiLineString",
                        coordinates: geometry.multi_line_string.line_string.map(
                            lineString => lineString.point.map(pointToGeoJsonCoordinate)
                        )
                    }
                });
            }
        }

        return features;
    }

    /**
     * Returns a promise with array of categories for a place. Retrieves categories names from metadata layer.
     *
     * @param {object} place A place object.
     * @param {string} placesRootTag Root tag for features.
     */
    async function getPlaceCategories(place, placesRootTag) {
        const names = [];
        for (const category of place.category) {
            const shortId = category.identifier.substring(category.identifier.lastIndexOf(":") + 1);
            const categoryItem = await getCategory(category.identifier, category.partition_name);
            if (categoryItem) {
                const combinedName = categoryItem.name.map(name => name.text.value).join(", ");
                if (placesRootTag) {
                    names.push(`${placesRootTag}|${combinedName} (${shortId})`);
                } else {
                    names.push(`${combinedName} (${shortId})`);
                }
            } else {
                names.push(shortId);
            }
        }
        return names;
    }

    /**
     * Returns an array with place names (place can have multiple names).
     *
     * @param {object} place A place object.
     */
    function getPlaceNames(place) {
        return place.name.map(name => name.text.value);
    }

    /**
     * Returns an array of nodes GeoJSON features.
     *
     * @param {Array} nodes Nodes array from decoded partition.
     */
    function processNodes(nodes) {
        return nodes.map((node, index) => ({
            type: "Feature",
            properties: {
                description: {
                    identifier: node.identifier
                },
                title: node.identifier,
                protobufRef: `node[${index}]`,
                radius: "5m",
                minRadius: "2px",
                featureTags: ["Road Topology & Geometry|Nodes"]
            },
            geometry: {
                type: "Point",
                coordinates: pointToGeoJsonCoordinate(node.geometry)
            }
        }));
    }

    /**
     * Returns an array of segments GeoJSON features.
     *
     * @param {Array} segments Segments array from decoded partition.
     */
    function processSegments(segments) {
        return segments.map((segment, index) =>
            ({
                type: "Feature",
                properties: {
                    description: {
                        identifier: segment.identifier
                    },
                    title: segment.identifier,
                    protobufRef: `segment[${index}]`,
                    width: "1px",
                    featureTags: ["Road Topology & Geometry|Segments"]
                },
                geometry: {
                    type: "LineString",
                    coordinates: segment.geometry.point.map(pointToGeoJsonCoordinate)
                }
            })
        );
    }

    /** Converts HMC point object to GeoJSON-compatible `[longitude, latitude]` pair. */
    function pointToGeoJsonCoordinate(point) {
        return [point.longitude, point.latitude];
    }
})();
