/*
 * Copyright (C) 2017-2020 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

/**
 * GeoJSON renderer plugin to visualize HMC extruded buildings data (from `3d-buildings` layer).
 */
(() => {
    "use strict";

    /** @type {RendererPlugin} */
    const plugin = {
        /**
         * Asynchronously generates GeoJSON from RIB Navigation Attributes data.
         *
         * @param {object} params.logger Logger to log information to console.
         * @param {object} params.layer All properties as well as dataUrl, name, partitioning and
         * originalConfiguration.
         * @param {object} params.decodeResult Input RIB Navigation Attributes data.
         * @param {object} params.mapUtils Various converters like `mercatorConverter`,
         * `heretileToQuad`, `hdmapCoordinateToLonLat`, etc.
         * @param {object} params.geometryDataProviders Various geometry data providers to map data.
         * @returns {Promise} A promise which resolves with generated GeoJSON data.
         */
        async getGeoJSONAsync(params) {
            const logger = params.logger;
            logger.level = 0;
            logger.info(`Creating GeoJSON for tile ${params.decodeResult.partitionId}`);
            const data = params.decodeResult.data;

            let features = [];

            if (data.location && data.place) {
                features = processBuildings(data);
            } else {
                logger.warn("Partition doesn't have location and place data.");
            }

            return {
                contentType: "application/vnd.geo+json; charset=utf-8",
                body: {
                    type: "FeatureCollection",
                    features
                }
            };
        }
    };

    return plugin;

    /**
     * Returns extruded buildings geometries.
     *
     * @param {object} data Decoded places partition data.
     */
    function processBuildings(data) {
        const features = [];
        const locations = new Map(data.location.map(item => [item.identifier, item]));

        for (let index = 0; index < data.place.length; index++) {
            const place = data.place[index];
            const location = locations.get(place.location_ref.identifier);
            if (location === undefined) {
                logger.warn(`Geometry for location "${place.location_ref.identifier}" not found`);
                return;
            }

            const geometry = location.geometry;
            if (geometry.multi_polygon === undefined && geometry.multi_line_string === undefined && geometry.multi_prism === undefined) {
                logger.warn(`Unsupported geometry type`, geometry);
                return;
            }

            const name = getPlaceNames(place).join(", ");
            const properties = {
                description: {
                    Identifier: place.identifier,
                    Name: name
                },
                title: `${name !== "" ? name + "<br>" : ""}${place.identifier}`,
                protobufRef: `place[${index}]`
            };

            if (geometry.multi_prism !== undefined) {
                const processRingWalls = (ring, z1, z2) => {
                    for (let i = 0; i < ring.point.length - 1; i++) {
                        const point1 = pointToGeoJsonCoordinate(ring.point[i]);
                        const point2 = pointToGeoJsonCoordinate(ring.point[i + 1]);
                        coordinates.push([[
                            [...point1, z1],
                            [...point2, z1],
                            [...point2, z2],
                            [...point1, z2],
                            [...point1, z1]
                        ]]);
                    }
                }

                const coordinates = [];

                geometry.multi_prism.prism.forEach(prism => {
                    const z1 = prism.ground_clearance;
                    const z2 = prism.ground_clearance + prism.height_clearance;
                    processRingWalls(prism.polygon.exterior_ring, z1, z2);

                    for (const ring of prism.polygon.interior_ring) {
                        processRingWalls(ring, z1, z2);
                    }
                });

                coordinates.push(...geometry.multi_prism.prism.map(prism => {
                    const z = prism.ground_clearance + prism.height_clearance;

                    const result = [
                        prism.polygon.exterior_ring.point.map(point => [...pointToGeoJsonCoordinate(point), z])
                    ];
                    prism.polygon.interior_ring.forEach(ring =>
                        result.push(ring.point.map(point => [...pointToGeoJsonCoordinate(point), z]))
                    );
                    return result;
                }));
                coordinates.push(...geometry.multi_prism.prism.map(prism => {
                    const z = prism.ground_clearance;

                    const result = [
                        prism.polygon.exterior_ring.point.map(point => [...pointToGeoJsonCoordinate(point), z])
                    ];
                    prism.polygon.interior_ring.forEach(ring =>
                        result.push(ring.point.map(point => [...pointToGeoJsonCoordinate(point), z]))
                    );
                    return result;
                }));

                features.push({
                    type: "Feature",
                    properties: {
                        ...properties,
                        "fill-opacity": 0.2
                    },
                    geometry: {
                        type: "MultiPolygon",
                        coordinates
                    }
                });
            }

            if (geometry.multi_polygon !== undefined) {
                features.push({
                    type: "Feature",
                    properties,
                    geometry: {
                        type: "MultiPolygon",
                        coordinates: geometry.multi_polygon.polygon.map(polygon => {
                            const result = [
                                polygon.exterior_ring.point.map(pointToGeoJsonCoordinate)
                            ];
                            polygon.interior_ring.forEach(ring =>
                                result.push(ring.point.map(pointToGeoJsonCoordinate))
                            );
                            return result;
                        })
                    }
                });
            }

            if (geometry.multi_line_string !== undefined) {
                features.push({
                    type: "Feature",
                    properties,
                    geometry: {
                        type: "MultiLineString",
                        coordinates: geometry.multi_line_string.line_string.map(
                            lineString => lineString.point.map(pointToGeoJsonCoordinate)
                        )
                    }
                });
            }
        }

        return features;
    }

    /**
     * Returns an array with place names (place can have multiple names).
     *
     * @param {object} place A place object.
     */
    function getPlaceNames(place) {
        return place.name.map(name => name.text.value);
    }

    /** Converts HMC point object to GeoJSON-compatible `[longitude, latitude]` pair. */
    function pointToGeoJsonCoordinate(point) {
        return [point.longitude, point.latitude];
    }
})();