/*
 * Copyright (C) 2017-2021 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

/**
 * Renderer plugin to convert HMC Places content JSON data to GeoJSON.
 * Retrieves categories from "places-metadata" layer of the catalog with source data.
 */
(() => {
    /** Logger. */
    let logger;
    /** Data provider of places metadata. */
    let metadataDataProvider;
    /** A map of locations, where a key is a location ID and value is a location object. */
    const locations = {};
    /** A map of categories, where a key is a category ID and value is a location object. */
    const categories = {};

    /** @type {RendererPlugin} */
    return {
        async getGeoJSONAsync(params) {
            logger = params.logger;
            const data = params.decodeResult.data;

            // Prepare a partition data provider to get place categories from "places-metadata" layer.
            metadataDataProvider = await params.getPartitionDataProvider("places-metadata");

            // Do preparations before creating features.
            await prepare(data);

            return {
                contentType: "application/geo+json",
                body: {
                    type: "FeatureCollection",
                    features: createFeatures(data)
                }
            };
        }
    };

    /**
     * Process decoded partition data. It converts needed arrays into maps in order to not use a
     * linear search, which is might be very slow for big arrays
     *
     * @param data Decoded partition data.
     */
    async function prepare(data) {
        // Convert an array locations into a map of locations.
        data.location.forEach(item => locations[item.identifier] = item);

        const metaPartition = {};
        // Convert an array of categories into a map of categories.
        await Promise.all(
            data.place.map(async place => await Promise.all(
                place.category.map(async item => {
                    const partition = item.partition_name;
                    if (metaPartition[partition]) {
                        return;
                    }

                    const metadataPromise = metadataDataProvider.getDataDecoded(partition);
                    metaPartition[partition] = metadataPromise;

                    const metadata = await metadataPromise;

                    metadata.category.forEach(item => categories[item.identifier] = item);
                }))
            )
        );
    }

    /**
     * Process decoded partition data to create GeoJSON features.
     *
     * @param data Decoded partition data.
     * @returns An array of GeoJSON features.
     */
    function createFeatures(data) {
        const features = [];

        for (let index = 0; index < data.place.length; index++) {
            const place = data.place[index];

            // Get the geometry
            const geometry = getGeometry(place, data);
            if (!geometry) {
                continue;
            }

            // Get an object of a place name
            const placeNameObj = place.name.find(item => item.name_type === "BASE_NAME") || place.name[0];

            features.push({
                type: "Feature",
                properties: {
                    name: placeNameObj && placeNameObj.text.value || "N/A",
                    categories: getCategories(place).join(", "),
                    protobufRef: `place[${index}]`,
                    radius: "3px"
                },
                geometry
            });
        }

        return features;
    }

    /**
     * Process a place to get the geometry object.
     *
     * @param place A place.
     * @returns A GeoJSON geometry object or undefined if location not found or geometry cannot be created.
     */
    function getGeometry(place) {
        const location = locations[place.location_ref.identifier];
        if (!location) {
            logger.warn(`Location for place ID "${place.identifier}" not found.`);
            return;
        }

        const geometry = {};
        switch (location.location_type) {
            case "POINT":
                geometry.type = "Point";
                geometry.coordinates = [
                    location.display_position.longitude,
                    location.display_position.latitude
                ];
                break;

            default:
                logger.warn(`Location type "${location.location_type}" is not yet supported for displaying`);
                return;
        }

        return geometry;
    }

    /**
     * Process a place to get its categories.
     *
     * @param place A place.
     * @returns An array of human readable categories.
     */
    function getCategories(place) {
        const placeCategories = [];
        place.category.forEach(item => {
            const category = categories[item.identifier];
            if (category) {
                placeCategories.push(category.name[0].text.value);
            }
        });

        return placeCategories;
    }
})();
