/*
 * Copyright (C) 2017-2020 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

/**
 * Renderer plugin to convert HMC content JSON data to GeoJSON.
 * Supports visualization of Road Topology and Geometry, Buildings and Cartography data.
 */
(() => {
    "use strict";

    /** Parameters for plugin. */
    let pluginParams;

    /** @type {RendererPlugin} */
    const plugin = {
        /**
         * Asynchronously generates GeoJSON from RIB Navigation Attributes data.
         *
         * @param {object} params.logger Logger to log information to console.
         * @param {object} params.layer All properties as well as dataUrl, name, partitioning and
         * originalConfiguration.
         * @param {object} params.decodeResult Input RIB Navigation Attributes data.
         * @param {object} params.mapUtils Various converters like `mercatorConverter`,
         * `heretileToQuad`, `hdmapCoordinateToLonLat`, etc.
         * @param {object} params.geometryDataProviders Various geometry data providers to map data.
         * @returns {Promise} A promise which resolves with generated GeoJSON data.
         */
        async getGeoJSONAsync(params) {
            pluginParams = params;

            const logger = params.logger;
            logger.level = 0;
            logger.info(`Creating GeoJSON for tile ${params.decodeResult.partitionId}`);
            const data = params.decodeResult.data;

            let features = [];

            if (data.node || data.segment) {
                features = features.concat(processSegments(data.segment || []));
                features = features.concat(processNodes(data.node || []));
            } else
                logger.error("Source data does not have HMC data. Make sure you are passing correct data to the renderer.");

            return {
                contentType: "application/vnd.geo+json; charset=utf-8",
                body: {
                    type: "FeatureCollection",
                    features
                }
            };
        }
    };

    return plugin;

    /**
     * Returns an array of nodes GeoJSON features.
     *
     * @param {Array} nodes Nodes array from decoded partition.
     */
    function processNodes(nodes) {
        return nodes.map((node, index) => ({
            type: "Feature",
            properties: {
                description: {
                    Identifier: node.identifier
                },
                title: node.identifier,
                protobufRef: `node[${index}]`,
                radius: "5m",
                minRadius: "2px",
                featureTags: ["Road Topology & Geometry|Nodes"]
            },
            geometry: {
                type: "Point",
                coordinates: pointToGeoJsonCoordinate(node.geometry)
            }
        }));
    }

    /**
     * Returns an array of segments GeoJSON features.
     *
     * @param {Array} segments Segments array from decoded partition.
     */
    function processSegments(segments) {
        const segments2 = [];
        for (const segment of segments) {
            const a = segment.geometry.point.find(p => p.z_level !== 0);
            if (a) {
                segments2.push(segment);
            }
        }

        return segments.map((segment, index) =>
            ({
                type: "Feature",
                properties: {
                    description: {
                        Identifier: segment.identifier
                    },
                    title: segment.identifier,
                    protobufRef: `segment[${index}]`,
                    width: "6px",
                    stroke: "red",
                    featureTags: ["Road Topology & Geometry|Segments"]
                },
                geometry: {
                    type: "LineString",
                    coordinates: segment.geometry.point.map(pointToGeoJsonCoordinate)
                }
            })
        );
    }

    /** Converts HMC point object to GeoJSON-compatible `[longitude, latitude]` pair. */
    function pointToGeoJsonCoordinate(point) {
        return [point.longitude, point.latitude, point.z_level * 20];
    }
})();
