/*
 * Copyright (C) 2017-2021 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

/**
 * Renderer plugin to convert JSON data to GeoJSON.
 *
 * Generated GeoJSON must be created according to GeoJSON Specification:
 * https://tools.ietf.org/html/rfc7946
 *
 * This template is an example of plugin to generate the GeoJSON object based on decoded partition
 * content (the same data structure as in Decoded Data panel).
 *
 * In order to use renderer plugin this file must be placed into project root directory and
 * parameter `externalRendererPluginName` must be passed into constructor:
 *
 * ```
 * const geoJsonDataSource = new GeoJsonDataSource({
 *    dataStore: {
 *        hrn: HRN.fromString("Your HRN string"),
 *        getBearerToken: tokenResolver.getBearerToken.bind(tokenResolver),
 *        layer: "Layer name"
 *    },
 *    externalRendererPluginName: "renderer.plugin.template"
 * });
 * ```
 *
 * Renderer plugin must return an object with one of these functions in it:
 *
 * ```
 * getGeoJSON: function (params) { ... }
 * getGeoJSONAsync: async function (params) { ... }
 * ```
 *
 * Both synchronous \`getGeoJSON\` and async \`getGeoJSONAsync\` functions return an object with
 * GeoJSON data.
 *
 * Only one of these functions must be declared, if both are present plugin will not be used.
 */
(() => {
    "use strict";

    let geometryDataProvider; // An instance of geometry data provider.
    let loggerInstance; // An instance of logger.

    // Thresholds and colors for different speed limit values.
    const MEDIUM_LIMIT_LOW_VALUE = 50;
    const MEDIUM_LIMIT_HIGH_VALUE = 80;
    const COLOR_SPEEDLIMIT_LOW = "#bf212f";
    const COLOR_SPEEDLIMIT_MEDIUM = "#e18c1e";
    const COLOR_SPEEDLIMIT_HIGH = "#27b376";
    const HOVER_SPEEDLIMIT_LOW = "#e52838";
    const HOVER_SPEEDLIMIT_MEDIUM = "#ffa423";
    const HOVER_SPEEDLIMIT_HIGH = "#2fd98f";

    /** @type {RendererPlugin} */
    const plugin = {
        /**
         * Asynchronously generates GeoJSON from RIB Navigation Attributes data.
         *
         * @param {object} params.logger Logger to log information to console.
         * @param {object} params.layer All properties as well as dataUrl, name, partitioning and
         * originalConfiguration.
         * @param {object} params.decodeResult Input RIB Navigation Attributes data.
         * @param {object} params.mapUtils Various converters like `mercatorConverter`,
         * `heretileToQuad`, `hdmapCoordinateToLonLat`, etc.
         * @param {object} params.geometryDataProviders Various geometry data providers to map data.
         * @returns {Promise} A promise which resolves with generated GeoJSON data.
         */
        async getGeoJSONAsync(params) {
            // Destructuring of `params` object.
            const { logger, decodeResult, geometryDataProviders } = params;

            // This is an async function so it must return a Promise.
            // Save an instance of logger so it can be used by other functions within the scope of
            // this plugin.
            loggerInstance = logger;

            // Get an instance of geometry data provider from `geometryDataProviders`.
            // Please note that this is an asynchronous call.
            geometryDataProvider = await geometryDataProviders.getSegmentAnchorDataProvider();

            // Validate input data.
            if (!decodeResult.data.speed_limit && !decodeResult.data.segment_anchor) {
                loggerInstance.error("Source data does not have speed limits data. Make sure you are passing correct data to the renderer.");
                throw new Error();
            }

            // Create result object.
            const result = {
                "type": "FeatureCollection",
                "features": await processSpeedLimits(decodeResult.data)
            };

            // Resolve promise with generated GeoJSON.
            return {
                contentType: "application/geo+json; charset=utf-8",
                body: result
            };
        }
    };

    return plugin;

    /**
     * Creates speed limits geometry.
     * Please note that this function is async.
     */
    async function processSpeedLimits(data) {
        const result = [];

        // Preliminary sort source data by speed limit value.
        data.speed_limit.sort((item1, item2) => item1.value < item2.value);

        // Iterate over speed limit items.
        let speedLimitItemIndex = 0;
        for (const speedLimitItem of data.speed_limit) {
            const style = getLineStyle(speedLimitItem);

            // Iterate over `SegmentAnchor`s of speed limit item.
            let segmentAnchorIndex = 0;
            for (const anchorIndex of speedLimitItem.segment_anchor_index) {
                // First, get actual `SegmentAnchor` from pool.
                const anchor = data.segment_anchor[anchorIndex];
                // Then, retrieve line coordinates from geometry data provider.
                const coordinates = await geometryDataProvider.getSegmentAnchorCoordinates(anchor);

                if (coordinates.length === 0) {
                    loggerInstance.warn(`Can't retrieve geometry for SegmentAnchor #${anchorIndex}`);
                }

                // Create a `LineString` feature for speed limit segment
                result.push({
                    type: "Feature",
                    properties: {
                        // `protobufRef` is a path pointing to original data property
                        protobufRef: `speed_limit[${speedLimitItemIndex}].segment_anchor_index[${segmentAnchorIndex}]`,
                        speedLimit: speedLimitItem.value,
                        style, // apply visual styles to feature
                        width: 4,
                        // `title` will be shown in tooltip when mouse is moved over the feature
                        title: speedLimitItem.value + " km/h"
                    },
                    geometry: {
                        type: "LineString",
                        coordinates
                    }
                });

                segmentAnchorIndex++;
            }
            speedLimitItemIndex++;
        };

        return result;
    }

    /**
     * Returns line style based on speed limit value.
     *
     * @param {object} speedLimitItem Speed limit item.
     * @returns {string} String with line color.
     */
    function getLineStyle(speedLimitItem) {
        if (speedLimitItem.value < MEDIUM_LIMIT_LOW_VALUE) {
            return { color: COLOR_SPEEDLIMIT_LOW, hoverColor: HOVER_SPEEDLIMIT_LOW };
        } else if (speedLimitItem.value < MEDIUM_LIMIT_HIGH_VALUE) {
            return { color: COLOR_SPEEDLIMIT_MEDIUM, hoverColor: HOVER_SPEEDLIMIT_MEDIUM };
        } else {
            return { color: COLOR_SPEEDLIMIT_HIGH, hoverColor: HOVER_SPEEDLIMIT_HIGH };
        }
    }
})();
