/*
 * Copyright (C) 2020-2021 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

(() => {
    "use strict";

    // Version of HMC catalog to retrieve road topology data.
    const HMC_CATALOG_VERSION = 1084;

    // Set to `true` to retrieve adjacent HMC topology tiles to eliminate gaps in traffic items
    // between adjacent traffic tiles.
    const RETRIEVE_ADJACENT_HMC_TILES = true;

    const R = 6378.137; // Radius of earth in KM
    const K = 2 * R * 1000; // Multiply by 1000 to convert to meters.
    const RADIAN = 180 / Math.PI;

    let geometryDataProvider; // An instance of geometry data provider.
    let loggerInstance; // An instance of logger.

    let mapUtils; // Utils to convert tile codes.
    let roadSegmentsProvider; // Road segments geometry provider.
    const adjacentTiles = {}; // Adjacent HMC tiles cache.

    /** @type {RendererPlugin} */
    const plugin = {
        /**
         * Asynchronously generates GeoJSON from RIB Navigation Attributes data.
         *
         * @param {object} params.logger Logger to log information to console.
         * @param {object} params.decodeResult Input RIB Navigation Attributes data.
         * @param {object} params.geometryDataProviders Various geometry data providers to map data.
         * @returns {Promise} A promise which resolves with generated GeoJSON data.
         */
        async getGeoJSONAsync(params) {
            const { logger, decodeResult, geometryDataProviders } = params;
            mapUtils = params.mapUtils;

            loggerInstance = logger;
            geometryDataProvider = await geometryDataProviders.getSegmentAnchorDataProvider({ catalogVersion: HMC_CATALOG_VERSION });
            roadSegmentsProvider = await geometryDataProviders.getRoadSegmentsDataProvider({ catalogVersion: HMC_CATALOG_VERSION });

            const adjacentMortonCodes = RETRIEVE_ADJACENT_HMC_TILES
                ? await retrieveAdjacentTiles(decodeResult.partitionId)
                : [];

            return {
                contentType: "application/geo+json; charset=utf-8",
                body: {
                    "type": "FeatureCollection",
                    "features": await processData(decodeResult, adjacentMortonCodes)
                }
            };
        }
    };

    return plugin;

    async function retrieveAdjacentTiles(partitionId) {
        const quad = mapUtils.heretileToQuad(+partitionId);
        const tile = mapUtils.quadKeyToTile(quad);
        const adjacentMortonCodes = [];

        for (let x = -1; x <= 1; x++) {
            for (let y = -1; y <= 1; y++) {
                const adjacentQuad = mapUtils.mercatorConverter.toQuad(tile.x + x, tile.y + y, tile.z);
                const adjacentHeretile = mapUtils.quadToHeretile(adjacentQuad);

                if (x === 0 && y === 0) {
                    continue;
                }
                adjacentMortonCodes.push(adjacentHeretile);

                if (adjacentTiles[adjacentHeretile] !== undefined) {
                    continue;
                }

                try {
                    const data = await roadSegmentsProvider.getRoadSegmentsForTile(adjacentHeretile);
                    adjacentTiles[adjacentHeretile] = data;
                } catch {
                    // Tile is out of bounds or has no data.
                }
            }
        }

        return adjacentMortonCodes;
    }

    /**
     * Creates traffic geometry.
     * @param decodeResult Decode result.
     * @param adjacentMortonCodes Adjacent tiles Morton codes.
     * @returns Array of traffic features.
     */
    async function processData(decodeResult, adjacentMortonCodes) {
        const { data, partitionId } = decodeResult;
        const result = [];

        for (let i = 0; i < data.items.length; i++) {
            const item = data.items[i];
            const anchor = {
                first_segment_start_offset: getStartOffset(item),
                last_segment_end_offset: getEndOffset(item),
                oriented_segment_ref: getSegmentRefs(item, partitionId, adjacentMortonCodes)
            };
            const segment = item.topology_segment;
            if (segment.is_first_segment_in_driving_direction === false && segment.topology_segment_id.length === 1) {
                if (segment.start_offset !== 0 && segment.end_offset !== 0) {
                    anchor.first_segment_start_offset.value = segment.end_offset;
                    anchor.last_segment_end_offset.value = 1 - segment.start_offset;
                } else if (segment.start_offset !== 0) {
                    anchor.first_segment_start_offset.value = 1 - segment.start_offset;
                } else if (segment.end_offset !== 0) {
                    anchor.last_segment_end_offset.value = segment.end_offset;
                }
            }

            if (segment.topology_segment_id.length > 1 && segment.end_offset !== 0) {
                const idLast = segment.topology_segment_id[segment.topology_segment_id.length - 1];
                const idNextToLast = segment.topology_segment_id[segment.topology_segment_id.length - 2];
                const anchorLast = {
                    first_segment_start_offset: { value: 0 },
                    last_segment_end_offset: { value: 1 },
                    oriented_segment_ref: [{
                        inverted: false,
                        segment_ref: {
                            partition_name: getSegmentPartition(idLast, adjacentMortonCodes) || partitionId,
                            identifier: `here:cm:segment:${idLast}`
                        }
                    }]
                };
                const anchorNextToLast = {
                    first_segment_start_offset: { value: 0 },
                    last_segment_end_offset: { value: 1 },
                    oriented_segment_ref: [{
                        inverted: false,
                        segment_ref: {
                            partition_name: getSegmentPartition(idNextToLast, adjacentMortonCodes) || partitionId,
                            identifier: `here:cm:segment:${idNextToLast}`
                        }
                    }]
                };
                const coordsLast = await geometryDataProvider.getSegmentAnchorCoordinates(anchorLast);
                const coordsNextToLast = await geometryDataProvider.getSegmentAnchorCoordinates(anchorNextToLast);

                if (coordsLast.length > 0 && coordsNextToLast.length > 0) {
                    const linesMerge = segment.is_first_segment_in_driving_direction
                        ? coordsLast[0][0] === coordsNextToLast[coordsNextToLast.length - 1][0] && coordsLast[0][1] === coordsNextToLast[coordsNextToLast.length - 1][1]
                        : coordsNextToLast[0][0] === coordsLast[coordsLast.length - 1][0] && coordsNextToLast[0][1] === coordsLast[coordsLast.length - 1][1];
                    if (!linesMerge) {
                        if (segment.is_first_segment_in_driving_direction) {
                            anchor.last_segment_end_offset.value = segment.end_offset;
                        } else {
                            anchor.last_segment_end_offset.value = 1 - segment.end_offset;
                        }
                    }
                }
            }

            const coordinates = await geometryDataProvider.getSegmentAnchorCoordinates(anchor);

            if (coordinates.length === 0) {
                console.warn(`Can't retrieve geometry for traffic item ${item.id}.  SegmentAnchor:`, anchor);
            } else {
                const features = getFeaturesForFlows(coordinates, item);
                features.forEach((feature, flowIndex) => {
                    feature.properties.protobufRef = `items[${i}].flow.segment[${flowIndex}]`;
                    // feature.properties.protobufRef = `items[${i}]`;
                    const itemTitle = features.length === 1
                        ? item.id
                        : `${item.id} (${flowIndex + 1} out of ${features.length})`;
                    feature.properties.title = itemTitle;
                    feature.properties.infoPopupTitle = itemTitle;
                    return feature;
                });
                result.push(...features);
            }
        }

        return result;
    }

    /**
     * Returns start offset of traffic item.
     * @param item Traffic item.
     * @returns Start offset of traffic item if exists, otherwise null.
     */
    function getStartOffset(item) {
        const segment = item.topology_segment;
        return segment.start_offset
            ? {
                value: segment.is_first_segment_in_driving_direction
                    ? segment.start_offset
                    : 1 - segment.start_offset
            } : null;
    }

    /**
     * Returns end offset of traffic item.
     * @param item Traffic item.
     * @returns End offset of traffic item if exists, otherwise null.
     */
    function getEndOffset(item) {
        const segment = item.topology_segment;

        return segment.end_offset
            ? {
                value: segment.is_first_segment_in_driving_direction
                    ? 1 - segment.end_offset
                    : segment.end_offset
            } : null;
    }

    /**
     * Returns topology segment.
     * @param item Traffic item.
     * @param partitionId Partition ID.
     * @param adjacentMortonCodes Adjacent tiles Morton codes.
     * @returns {object} Topology segment.
     */
    function getSegmentRefs(item, partitionId, adjacentMortonCodes) {
        return item.topology_segment.topology_segment_id.map((id, index) => {
            return {
                inverted: index === 0 ? !item.topology_segment.is_first_segment_in_driving_direction : false,
                segment_ref: {
                    partition_name: getSegmentPartition(id, adjacentMortonCodes) || partitionId,
                    identifier: `here:cm:segment:${id}`
                }
            };
        })
    }

    /**
     * Finds segment in adjacent partitions and if found returns its id. If not found, returns `undefined`.
     * @param {number} segmentId
     * @param {array} adjacentMortonCodes
     */
    function getSegmentPartition(segmentId, adjacentMortonCodes) {
        let result = undefined;

        for (const mortonCode of adjacentMortonCodes) {
            const tile = adjacentTiles[mortonCode];
            const found = tile !== undefined && tile.find(item => item.segmentId === segmentId) !== undefined;
            if (found) {
                result = mortonCode.toString();
                break;
            }
        }

        return result;
    }

    /**
     * Returns distance between two points in meters.
     * @param pointA An array of type [latitude, longitude].
     * @param pointB An array of type [latitude, longitude].
     * @returns {number} Distance between two points in meters.
     */
    function getLength(pointA, pointB) {
        const [lat1, lon1] = pointA;
        const [lat2, lon2] = pointB;
        const dLat = (lat2 - lat1) / RADIAN;
        const dLon = (lon2 - lon1) / RADIAN;
        const a = Math.sin(dLat / 2) ** 2 +
            Math.cos(lat1 / RADIAN) * Math.cos(lat2 / RADIAN) * Math.sin(dLon / 2) ** 2;
        return K * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
    }

    /**
     * Returns length of line in meters.
     * @param coordinates An array of points.
     * @returns {number} Length of line in meters.
     */
    function getTotalLength(coordinates) {
        let result = 0;
        for (let i = 0; i < coordinates.length - 1; i++) {
            result += getLength(coordinates[i], coordinates[i + 1]);
        }
        return result;
    }

    /**
     * Returns coordinates of a point C in a segment AB that divide this segment in a given ratio.
     * @param pointA An array of type [latitude, longitude].
     * @param pointB An array of type [latitude, longitude].
     * @param coeff Ratio value to divide a given segment AB.
     * @returns {[number, number]} Coordinates of a point C in a segment AB that divide this
     * segment in a given ratio.
     *
     * For example:
     * A---C-----B
     *
     * Point C divides segment AB. Ratio value = AC/CB = 3/5.
     */
    function getSegmentPartCoordinates(pointA, pointB, coeff) {
        const [x1, y1] = pointA;
        const [x2, y2] = pointB;
        const x = (x1 + coeff * x2) / (1 + coeff);
        const y = (y1 + coeff * y2) / (1 + coeff);
        return [x, y];
    }

    /**
     * Returns an array of features that represents traffic flow items.
     * @param coordinates An array of coordinates.
     * @param trafficItem Traffic item.
     * @returns {array} An array of features.
     */
    function getFeaturesForFlows(coordinates, trafficItem) {
        const flow = trafficItem.flow;
        const totalLength = getTotalLength(coordinates);

        // If flow contains only one segment just return feature for it.
        if (flow.segment.length === 1) {
            return [
                getFeature(coordinates, totalLength, flow.segment[0], flow.confidence)
            ];
        }

        const flowSegments = [];

        const lengthBreakpoints = [];
        for (let i = 1; i < flow.segment.length; i++) {
            lengthBreakpoints.push(flow.segment[i].start_offset * totalLength);
        }

        let startPoint = coordinates[0];
        let endPoint;
        let segmentCoordinates = [startPoint];
        let accumulatedLength = 0;
        let breakPointIndex = 0;
        for (let i = 1; i < coordinates.length; i++) {
            endPoint = coordinates[i];
            const lineLength = getLength(startPoint, endPoint);
            if (accumulatedLength + lineLength > lengthBreakpoints[breakPointIndex]) {
                const secondPartLength = accumulatedLength + lineLength - lengthBreakpoints[breakPointIndex];
                const firstPartLength = lineLength - secondPartLength;
                const coeff = firstPartLength / secondPartLength;
                const segmentPartCoordinates = getSegmentPartCoordinates(startPoint, endPoint, coeff);
                segmentCoordinates.push(segmentPartCoordinates);
                accumulatedLength += firstPartLength;

                flowSegments.push({
                    coordinates: segmentCoordinates,
                    length: accumulatedLength - (lengthBreakpoints[breakPointIndex - 1] || 0)
                });

                startPoint = segmentPartCoordinates;
                segmentCoordinates = [startPoint];
                breakPointIndex++;
                i--;
            } else {
                segmentCoordinates.push(endPoint);
                accumulatedLength += lineLength;
                startPoint = endPoint;

                if (accumulatedLength === lengthBreakpoints[breakPointIndex]) {
                    flowSegments.push({
                        coordinates: segmentCoordinates,
                        length: accumulatedLength - (lengthBreakpoints[breakPointIndex - 1] || 0)
                    });
                    segmentCoordinates = [startPoint];
                    breakPointIndex++;
                }
            }
        }

        if (segmentCoordinates.length > 1) {
            flowSegments.push({
                coordinates: segmentCoordinates,
                length: accumulatedLength - lengthBreakpoints[lengthBreakpoints.length - 1]
            });
        }

        return flowSegments.map((item, i) => {
            return getFeature(item.coordinates, item.length, flow.segment[i], flow.confidence);
        });
    }

    /**
     * Creates a GeoJson LineString feature.
     * @param coordinates An array of coordinates.
     * @param length Length of the line in meters.
     * @param flowSegment Segment that represents a single flow.
     * @param confidence Confidence value.
     * @returns {object} GeoJson LineString feature.
     */
    function getFeature(coordinates, length, flowSegment, confidence) {
        const color = getFlowColor(flowSegment);
        return {
            type: "Feature",
            properties: {
                description: {
                    "Length": length.toFixed(1) + " m",
                    "Average Speed Capped": `${flowSegment.speed ? flowSegment.speed.average_speed_capped_kph : '---'} km/h`,
                    "Average Speed": `${flowSegment.speed ? flowSegment.speed.average_speed_kph : '---'} km/h`,
                    "Free Flow Speed": `${flowSegment.speed ? flowSegment.speed.free_flow_speed_kph : '---'} km/h`,
                    "Confidence Value": confidence,
                    "Jam Factor": flowSegment.jam_factor
                },

                width: 2,
                style: {
                    color,
                    opacity: 0.5,
                    selectedColor: color,
                    selectedOpacity: 0.85,
                    hoverColor: color,
                    hoverOpacity:0.9
                },
            },
            geometry: {
                type: "LineString",
                coordinates: getCoordinatesWithOffset(coordinates)
            }
        };
    }

    /**
     * Returns color for traffic flow segment.
     *
     * @param segment Source segment.
     * @returns {string} Hex string with a color of segment.
     */
    function getFlowColor(segment) {
        // 1 - "CLOSED"
        if (segment.traversability_status === 1) {
            return "#000000";
        }
        // 2 - "ROAD_NOT_ROUTABLE"
        if (segment.traversability_status === 2) {
            return "#000000";
        }

        const jamFactor = segment.jam_factor !== undefined ? segment.jam_factor : 0;

        if (jamFactor >= 0 && jamFactor < 4) {
            return "#89C690"; // green
        } else if (jamFactor >= 4 && jamFactor < 6.3) {
            return "#FED500"; // yellow
        } else if (jamFactor >= 6.3 && jamFactor < 8) {
            return "#FE9925"; // orange
        } else if (jamFactor >= 8 && jamFactor < 10) {
            return "#E30F39"; // red
        }

        return "#000000";
    }

    /**
     * Returns new array of coordinates with offset.
     * @param coordinates An array of coordinates.
     * @returns {array} An array of coordinates with offset.
     */
    function getCoordinatesWithOffset(coordinates) {
        const normals = getPolylineNormals(coordinates);
        // Coefficient to make the proper distance between the lanes.
        const offsetMultiplier = 0.000016;

        return coordinates.map((point, i) => {
            const normal = normals[i];
            return [
                point[0] - normal[0] * offsetMultiplier,
                point[1] - normal[1] * offsetMultiplier
            ];
        });
    }

    /**
     * For the given coordinates array, produces a new array of the same length with normal
     * information for each point.
     * @param line An array of coordinates.
     * @returns {array} An array with normal information for each point.
     */
    function getPolylineNormals(line) {
        let normal, vectorA = [0, 0], vectorB = [0, 0], miter = [0, 0];
        const result = [];
        for (let i = 1, N = line.length; i < N; i++) {
            const prevPoint = line[i - 1];
            const point = line[i];
            const nextPoint = line[i + 1];

            getNormalizedVectorFromPoints(vectorA, point, prevPoint);
            if (normal === undefined) {
                normal = [0, 0];
                getPerpendicular(normal, vectorA);
            }

            if (i === 1) {
                result.push([...normal]);
            }

            if (nextPoint === undefined) {
                getPerpendicular(normal, vectorA);
                result.push([...normal]);
            } else {
                getNormalizedVectorFromPoints(vectorB, nextPoint, point);
                computeMiter(miter, vectorA, vectorB);
                result.push([...miter]);
            }
        }
        return result;
    }

    /**
     * Returns normalized vector for given points.
     * @param result The resulting vector.
     * @param pointA Start point of vector.
     * @param pointB End point of vector.
     * @returns {array} The resulting vector.
     */
    function getNormalizedVectorFromPoints(result, pointA, pointB) {
        subtractVectors(result, pointA, pointB);
        return normalizeVector(result, result);
    }

    /**
     * Builds a perpendicular vector for a given vector and saves it in first parameter.
     * @param result The resulting vector.
     * @param vector Vector.
     */
    function getPerpendicular(result, vector) {
        result[0] = -vector[1];
        result[1] = vector[0];
    }

    /**
     * Computes miter join for two vectors and saves it in first parameter.
     * @param miter Miter join for two vectors.
     * @param vectorA Vector A.
     * @param vectorB Vector B.
     */
    function computeMiter(miter, vectorA, vectorB) {
        const sum = [];
        addVectors(sum, vectorA, vectorB);
        normalizeVector(sum, sum);
        miter[0] = -sum[1];
        miter[1] = sum[0];
    }

    /**
     * Add two vectors and saves result in first parameter.
     * @param result The resulting vector.
     * @param vectorA Vector A.
     * @param vectorB Vector B.
     * @returns {array} The result of adding two vectors.
     */
    function addVectors(result, vectorA, vectorB) {
        result[0] = vectorA[0] + vectorB[0];
        result[1] = vectorA[1] + vectorB[1];
        return result;
    }

    /**
     * Subtracts two vectors and saves result in first parameter.
     * @param result The resulting vector.
     * @param vectorA Vector A.
     * @param vectorB Vector B.
     * @returns {array} The result of subtracting two vectors.
     */
    function subtractVectors(result, vectorA, vectorB) {
        result[0] = vectorA[0] - vectorB[0];
        result[1] = vectorA[1] - vectorB[1];
        return result;
    }

    /**
     * Normalizes given vector and saves result in first parameter.
     * @param result The resulting vector.
     * @param vector Given vector.
     * @returns {array} The normalized vector.
     */
    function normalizeVector(result, vector) {
        const [x, y] = vector;
        const length = Math.sqrt(x ** 2 + y ** 2);
        if (length === 0) {
            return result;
        }
        result[0] = x / length;
        result[1] = y / length;
        return result;
    }
})();
