/*
 * Copyright (C) 2019-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

import { DataInspector } from "@here/interactive-mapview-ui";
import { InteractiveIntersectionAgent } from "@here/interactive-datasource-protocol";
import { GeoJsonData, GeoJsonDataSource, Feature, LineString } from "@here/olp-geojson-datasource";
import { useToken, locale, showLanguageSelector, geoJsonMapCenter, lookupServiceEnvironment, performanceGeoJsonTile } from "../config";
import { GeoBox, GeoCoordinates, TileKey } from "@here/harp-geoutils";
import { bearerTokenProvider, InformationPanel } from "./common";

class GeoJsonPerformanceApplication {
    static readonly MAX_POINTS_IN_ROW = 500;
    static readonly LINES_POINT_COUNT = 500;
    static readonly MAX_POLYGONS_IN_ROW = 500;
    static readonly POLYGON_POINTS_COUNT = 32;

    /** An instance of [[DataInspector]] UI widget. */
    private m_dataInspector: DataInspector;

    /** An instance of GeoJsonDataSource class. */
    private m_geoJsonDataSource: GeoJsonDataSource;

    /** An instance of InfoPanel class. */
    private m_infoPanel: InformationPanel;

    constructor() {
        this.m_dataInspector = new DataInspector(
            {
                mapView: {
                    theme: "../resources/normal.reduced.night.json",
                    decoder: {
                        url: "../decoder.bundle.js"
                    },
                    defaultZoomLevel: 12
                },
                monacoEditorWorkersBasePath: "../monaco-editor",
                translationsBasePath: "../resources/i18n/",
                streamSaverBasePath: "../resources/mitm/",
                defaultLocale: locale || undefined,
                getBearerToken: useToken ? bearerTokenProvider : undefined,
                widgets: {
                    authForm: !useToken ? {
                        tokenForm: false,
                        accessKeyForm: true,
                        localAuthForm: true
                    } : undefined
                },
                toolbar: {
                    languageSelector: showLanguageSelector
                },
                lookupServiceEnvironment
            },
            document.getElementById("map-here") as HTMLElement
        );

        if (this.m_dataInspector.mapView === undefined) {
            throw new Error("MapView was not initialized.");
        }

        // Open control panel to pay attention of the user on authorization form in case if not
        // logged in.
        if (
            this.m_dataInspector.toolBar !== undefined &&
            this.m_dataInspector.authForm !== undefined &&
            this.m_dataInspector.authForm.isDisabled === false
        ) {
            const leftPanelBtn = this.m_dataInspector.toolBar.getLeftPanelBtn();
            if (leftPanelBtn) {
                leftPanelBtn.switchOn();
            }
        }

        const mapView = this.m_dataInspector.mapView;
        mapView.geoCenter = new GeoCoordinates(
            geoJsonMapCenter.latitude, geoJsonMapCenter.longitude
        );

        // geoJSON Data source
        this.m_geoJsonDataSource = new GeoJsonDataSource({
            intersectionAgent: new InteractiveIntersectionAgent(mapView)
        });

        mapView.addDataSource(this.m_geoJsonDataSource);

        this.m_infoPanel = new InformationPanel("");

        // Adds submit handler for user input form.
        this.initializeUserInputForm();
    }

    /**
     * Adds submit handler of user input form where user can change some parameters to test
     * rendering performance.
     */
    private initializeUserInputForm(): void {
        const form = document.getElementById("userInputForm") as HTMLFormElement;
        if (form === null) {
            return;
        }
        form.onsubmit = this.userInputFormHandler.bind(this);
    }

    /**
     * Gathers user input, generate random GeoJson object based on entered data, render it.
     */
    private userInputFormHandler(event: Event): boolean {
        const formData = new FormData(event.target as HTMLFormElement);

        const points = parseInt(formData.get("points") as string) || 0;
        const radius = parseInt(formData.get("radius") as string) || 0;
        const lines = parseInt(formData.get("lines") as string) || 0;
        const polygons = parseInt(formData.get("polygons") as string) || 0;

        const bbox = this.m_geoJsonDataSource.getTilingScheme().getGeoBox(
            TileKey.fromMortonCode(performanceGeoJsonTile)
        );

        const features: Feature[] = [];
        features.push(...this.generatePoints(points, bbox, radius));
        features.push(...this.generateLines(lines, bbox));
        features.push(...this.generatePolygons(polygons, bbox));

        const dataToRender: GeoJsonData = {
            type: "FeatureCollection",
            features
        };
        const startRenderingTimestamp = Date.now();
        this.m_geoJsonDataSource.renderGeoJson(dataToRender).then(() => {
            const finishRenderingTimestamp = Date.now();
            this.m_infoPanel.innerText = `${finishRenderingTimestamp - startRenderingTimestamp}`;
        });

        // Prevent default submit behavior.
        return false;
    }

    /**
     * @param count of points to generate.
     * @param bbox object with tile coordinates.
     * @param radius of points.
     * @returns array of generated points.
     */
    private generatePoints(count: number, bbox: GeoBox, radius: number): Feature[] {
        const result: Feature[] = [];

        if (count <= 0) {
            return result;
        }

        const stepX = bbox.latitudeSpan / GeoJsonPerformanceApplication.MAX_POINTS_IN_ROW;
        const stepY = bbox.longitudeSpan / Math.ceil(count / GeoJsonPerformanceApplication.MAX_POINTS_IN_ROW);

        for (let i = 0; i < count; i++) {
            const x = i % GeoJsonPerformanceApplication.MAX_POINTS_IN_ROW;
            const y = Math.floor(i / GeoJsonPerformanceApplication.MAX_POINTS_IN_ROW);
            const point: Feature = {
                "type": "Feature",
                "properties": {
                    "radius": radius,
                    "fill": "#00a0a0",
                    "stroke": "#ff0000",
                    "stroke-width": 4,
                    "fill-opacity": 0.7,
                    "stroke-opacity": 1
                },
                "geometry": {
                    "type": "Point",
                    "coordinates": [bbox.east + stepX * x, bbox.north + stepY * y]
                }
            };
            result.push(point);
        }

        return result;
    }

    /**
     * @param count of lines to generate.
     * @param bbox object with tile coordinates.
     * @returns array of generated lines.
     */
    private generateLines(count: number, bbox: GeoBox): Feature[] {
        const result: Feature[] = [];

        if (count <= 0) {
            return result;
        }

        const stepX = bbox.longitudeSpan / GeoJsonPerformanceApplication.LINES_POINT_COUNT;
        const stepY = bbox.latitudeSpan / count;

        for (let y = bbox.south; y < bbox.north; y += stepY) {
            const line: Feature = {
                "type": "Feature",
                "properties": {
                    "color": "#00dddd",
                    "stroke-width": 3
                },
                "geometry": {
                    "type": "LineString",
                    "coordinates": []
                }
            };
            for (let x = bbox.west; x < bbox.east; x += stepX) {
                (line.geometry as LineString).coordinates.push([x, y + Math.random() * stepY / 2]);
            }
            result.push(line);
        }

        return result;
    }

    /**
     * @param count of polygons to generate.
     * @param bbox object with tile coordinates.
     * @returns array of generated polygons.
     */
    private generatePolygons(count: number, bbox: GeoBox): Feature[] {
        const result: Feature[] = [];

        if (count <= 0) {
            return result;
        }

        const stepX = bbox.longitudeSpan / Math.ceil(count / GeoJsonPerformanceApplication.MAX_POLYGONS_IN_ROW);
        const stepY = bbox.latitudeSpan / GeoJsonPerformanceApplication.MAX_POLYGONS_IN_ROW;
        const radius = Math.min(stepX, stepY) * 0.45;

        for (let i = 0; i < count; i++) {
            const x = bbox.north + stepX * Math.floor(i / GeoJsonPerformanceApplication.MAX_POLYGONS_IN_ROW);
            const y = bbox.east + stepY * (i % GeoJsonPerformanceApplication.MAX_POLYGONS_IN_ROW);
            const polygon: Feature = {
                "type": "Feature",
                "properties": {
                    "color": "#00dddd",
                    "stroke-width": 3
                },
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [
                        this.getCircleCoordinates(x, y, radius),
                        this.getCircleCoordinates(
                            x + (Math.random() - 0.5) * radius * 0.4,
                            y + (Math.random() - 0.5) * radius * 0.4,
                            radius / 2
                        ).reverse()
                    ]
                }
            };
            result.push(polygon);
        }

        return result;
    }

    private getCircleCoordinates(
        centerX: number,
        centerY: number,
        radius: number
    ): number[][] {
        const result = [];
        const step = Math.PI * 2 / GeoJsonPerformanceApplication.POLYGON_POINTS_COUNT;

        for (let a = 0; a < Math.PI * 2; a += step) {
            const x = Math.sin(a) * radius + centerX;
            const y = Math.cos(a) * radius + centerY;

            result.push([y, x]);
        }
        result.push(result[0], result[1]);

        return result;
    }
}

new GeoJsonPerformanceApplication();