/*
 * Copyright (C) 2019-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

import { DataInspector } from "@here/interactive-mapview-ui";
import { InteractiveDataSource } from "@here/interactive-datasource-protocol";
import { MapViewEventNames, RenderEvent } from "@here/harp-mapview";
import { useToken, locale, showLanguageSelector, ribHostHrn, lookupServiceEnvironment, performanceRibTiles } from "../config";
import { bearerTokenProvider, InformationPanel } from "./common";
import * as THREE from "three";

export class RibPerformanceApplication {
    /** An instance of [[DataInspector]] UI widget. */
    private m_dataInspector: DataInspector;

    /** An instance of InfoPanel class. */
    private m_infoPanel: InformationPanel;

    /** A count of tiles to render. */
    private m_numberOfTilesToRender: number | undefined;

    /** Map of timestamps when tile starts rendering for tile morton keys. */
    private m_startTimestampsMap: Map<number, number> = new Map();

    /** Map of timestamps when tile finishes rendering for tile morton keys. */
    private m_finishTimestampsMap: Map<number, number> = new Map();

    constructor(ribLayer: string) {
        this.m_dataInspector = new DataInspector(
            {
                mapView: {
                    defaultZoomLevel: 9,
                    theme: "../resources/normal.reduced.night.json",
                    decoder: {
                        url: "../decoder.bundle.js"
                    }
                },
                monacoEditorWorkersBasePath: "../monaco-editor",
                translationsBasePath: "../resources/i18n/",
                streamSaverBasePath: "../resources/mitm/",
                defaultLocale: locale || undefined,
                getBearerToken: useToken ? bearerTokenProvider : undefined,
                widgets: {
                    authForm: !useToken ? {
                        tokenForm: false,
                        accessKeyForm: true,
                        localAuthForm: true
                    } : undefined
                },
                toolbar: {
                    languageSelector: showLanguageSelector
                },
                lookupServiceEnvironment
            },
            document.getElementById("map-here") as HTMLElement,
            {
                interactiveDataSource: {
                    hrn: ribHostHrn,
                    layer: ribLayer
                },
                enableCoverage: true,
                enableFiltering: true,
                enableInfoPanel: true
            }
        );

        if (this.m_dataInspector.mapView === undefined) {
            throw new Error("MapView was not initialized.");
        }

        this.m_infoPanel = new InformationPanel("");

        this.m_dataInspector.mapView.addEventListener(MapViewEventNames.DataSourceConnect, (event: RenderEvent) => {
            const dataSource = this.m_dataInspector.interactiveDataSource;
            if (dataSource !== undefined && event.dataSourceName === dataSource.name) {
                dataSource.addEventListener(InteractiveDataSource.EVENT_TILE_SELECT, (e: THREE.Event) => {
                    this.m_startTimestampsMap.set(e.tileKey.m_mortonCode, Date.now());
                });
                dataSource.addEventListener(InteractiveDataSource.EVENT_TILE_DECODED, (e: THREE.Event) => {
                    this.m_finishTimestampsMap.set(e.tileKey.m_mortonCode, Date.now());
                    this.updateInfoPanel();
                });
            }
        });

        // Open control panel to pay attention of the user on authorization form in case if not
        // logged in.
        if (
            this.m_dataInspector.toolBar !== undefined &&
            this.m_dataInspector.authForm !== undefined &&
            this.m_dataInspector.authForm.isDisabled === false
        ) {
            const leftPanelBtn = this.m_dataInspector.toolBar.getLeftPanelBtn();
            if (leftPanelBtn) {
                leftPanelBtn.switchOn();
            }
        }

        // Adds submit handler for user input form.
        this.initializeUserInputForm();
    }

    /**
     * Adds submit handler of user input form where user can change some parameters to test
     * rendering performance.
     */
    private initializeUserInputForm(): void {
        const form = document.getElementById("userInputForm") as HTMLFormElement;
        if (form === null) {
            return;
        }
        form.onsubmit = this.userInputFormHandler.bind(this);
    }

    /**
     * Gathers user input, generate random GeoJson object based on entered data, render it.
     */
    private userInputFormHandler(event: Event): boolean {
        this.clearData();

        const formData = new FormData(event.target as HTMLFormElement);
        this.m_numberOfTilesToRender = Number.parseInt(formData.get("numberOfTiles") as string, 10) || 1;

        const tilesToSelect = performanceRibTiles.slice(0, this.m_numberOfTilesToRender);
        tilesToSelect.forEach(t => {
            if (this.m_dataInspector.interactiveDataSource) {
                this.m_dataInspector.interactiveDataSource.selectTile(t);
            }
        });

        // Prevent default submit behavior.
        return false;
    }

    /**
     * Reset value for calculated properties.
     * Reset text inside InfoPanel instance.
     * Reset selected tile.
     */
    private clearData() {
        this.m_numberOfTilesToRender = undefined;
        this.m_startTimestampsMap.clear();
        this.m_finishTimestampsMap.clear();
        this.m_infoPanel.innerText = "";
        if (this.m_dataInspector.interactiveDataSource) {
            this.m_dataInspector.interactiveDataSource.clearSelectedTiles();
        }
    }

    /**
     * Calculate the difference between start rendering timestamp of first tile and finish
     * rendering timestamp of last tile. Display the calculated value inside InfoPanel element.
     */
    private updateInfoPanel() {
        if (this.m_numberOfTilesToRender === this.m_finishTimestampsMap.size) {
            const minStartTimestamp = Math.min(...this.m_startTimestampsMap.values());
            const maxFinishTimestamp = Math.max(...this.m_finishTimestampsMap.values());
            this.m_infoPanel.innerText = `${maxFinishTimestamp - minStartTimestamp}`;
        }
    }
}