/*
 * Copyright (C) 2017-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

import { GeoCoordinates } from "@here/harp-geoutils";
import { MapView, MapViewEventNames, CopyrightElementHandler } from "@here/harp-mapview";
import { VectorTileDataSource } from "@here/harp-vectortile-datasource";
import {
    AccessTokenResolver,
    ControlPanel,
    DEFAULT_COPYRIGHT_CHINA, DEFAULT_COPYRIGHT,
    InteractiveWidget,
    TileControlHandler,
    UserAuthForm
} from "@here/interactive-mapview-ui";
import { EnvironmentName } from "@here/olp-sdk-core";
import { geoCenter, omvDataLayer, omvHrn, lookupServiceEnvironment } from "../config";
import {
    BaseMapDataProvider,
    LookupEnvironment,
    MapControlHandler,
} from "@here/interactive-datasource-protocol";
import "@here/interactive-mapview-ui/resources/interactiveFilteringControl.css";
import "@here/interactive-mapview-ui/resources/onOffSwitch.css";
import { OlpCopyrightProvider } from "@here/harp-olp-utils";

/**
 * Initializes MapView instance
 *
 * @param canvas Canvas element to be used in MapView
 */
export function initializeMapView(canvas: HTMLCanvasElement): MapView {
    canvas.style.height = "100%";

    const mapView = new MapView({
        canvas,
        theme: "../resources/normal.reduced.night.json",
        decoderUrl: "../decoder.bundle.js",
        maxZoomLevel: isChinaEnv(lookupServiceEnvironment)? 20 : 24,
        minZoomLevel: isChinaEnv(lookupServiceEnvironment) ? 4 : 1,
        maxVisibleDataSourceTiles: Infinity
    });

    // center the camera somewhere around Berlin geo locations
    mapView.geoCenter = new GeoCoordinates(geoCenter.latitude, geoCenter.longitude);

    // resize the mapView to maximum
    mapView.resize(window.innerWidth, window.innerHeight);

    const mapContainer = document.getElementById("map-container") as HTMLDivElement;

    // react on resize events
    window.addEventListener(
        "resize",
        () => mapView.resize(mapContainer.offsetWidth, mapContainer.offsetHeight)
    );

    // react on changing of DOM
    const target = document.getElementById("inner-container");
    if (target) {
        new MutationObserver(
            () => mapView.resize(mapContainer.offsetWidth, mapContainer.offsetHeight)
        ).observe(target, { childList: true }); // monitor only for addition/removal of child nodes
    }
    initCopyrightHandler(mapView);
    return mapView;
}

/**
 * Initializes copyright handler and installs it
 *
 * @param mapView instance of MapView to attach handler to
 */
export function initCopyrightHandler(mapView: MapView) {
    const mapContainer = document.getElementById("map-container") as HTMLDivElement;
    const copyrightScaleContainer = document.createElement("div");
    copyrightScaleContainer.setAttribute("class", "map-scale-copyright");
    const copyrightContainer = document.createElement("div");
    copyrightContainer.setAttribute("class", "map-copyright");

    copyrightScaleContainer.appendChild(copyrightContainer);
    mapContainer.appendChild(copyrightScaleContainer);
    const handler = CopyrightElementHandler.install(copyrightContainer, mapView);
    handler.setStaticCopyightInfo(
        isChinaEnv(lookupServiceEnvironment) ? DEFAULT_COPYRIGHT_CHINA : DEFAULT_COPYRIGHT
    );
}

/**
 * Info panel.
 */
export class InformationPanel {

    /** Root element. */
    readonly element: HTMLDivElement = document.createElement("div");

    /**
     * Default constructor.
     *
     * @param m_infoText Info text.
     * @param m_isHTML `true` when info text is html.
     */
    constructor(private m_infoText: string, private m_isHTML?: boolean) {
        this.element.classList.add("info-panel");
        document.body.appendChild(this.element);
        this.setDefaultText();
    }

    /**
     * Sets inner text.
     *
     * @param content Content.
     */
    set innerText(content: string) {
        this.element.innerText = content;
    }

    /**
     * Sets inner HTML.
     *
     * @param content Content.
     */
    set innerHTML(content: string) {
        this.element.innerHTML = content;
    }

    /** Returns class list. */
    get classList(): DOMTokenList {
        return this.element.classList;
    }

    /**
     * Sets an error message.
     *
     * @param errorMessage Error message.
     */
    setError(errorMessage: string) {
        this.element.classList.add("error");
        this.element.innerText = errorMessage;
    }

    /** Sets default text. */
    setDefaultText() {
        this.element.classList.remove("error");
        if (this.m_isHTML === true) {
            this.element.innerHTML = this.m_infoText;
        } else {
            this.element.innerText = this.m_infoText;
        }
    }
}

/**
 * Adds an event listener for the [[MapViewEventNames.DataSourceConnect]] event to
 * show proper warning message in case of data source connection failure.
 */
export function addConnectionListener(authForm: UserAuthForm, mapView: MapView): void {
    mapView.addEventListener(MapViewEventNames.DataSourceConnect, (event: any) => {
        if (event.error !== undefined) {
            authForm.showWarningNotice(event.dataSourceName, event.error.message);
        }
    });
}

/**
 * Creates basemap (OMV) datasource.
 *
 * @param getToken Callback to retrieve authentication token.
 * @param environment Environment for the Lookup Service.
 */
export async function createBaseMapDataSource(
    getToken: () => Promise<string>,
    environment: EnvironmentName = lookupServiceEnvironment
): Promise<VectorTileDataSource> {
    const dataProvider = new BaseMapDataProvider({
        hrn: omvHrn,
        layer: omvDataLayer,
        getToken,
        environment
    });

    await dataProvider.connect();
    const version = await dataProvider.catalog!.getLatestVersion();
    // TODO: For CN environments, temporarily use hardcoded copyrights, don't retrieve them from OMV
    const copyrightProvider = isChinaEnv(environment)
        ? undefined
        : new OlpCopyrightProvider({
            hrn: omvHrn,
            getToken,
            version,
            environment
        });
    return new VectorTileDataSource({
        dataProvider,
        styleSetName: "tilezen",
        maxDataLevel: 14,
        copyrightProvider,
        copyrightInfo: isChinaEnv(environment) ? DEFAULT_COPYRIGHT_CHINA :
            DEFAULT_COPYRIGHT,
        //TODO IP: hack to fix vertical line artifacts on OMV
        // https://saeljira.it.here.com/browse/HARP-12145
        roundUpCoordinatesIfNeeded: true,
    });
}

/**
 * Adds various fixes for China base map: 1)  removes country border outlines. 2) add special
 * datasource to cut off labels outside of country border.
 *
 * @param mapView An instance of  MapView.
 * @param dataProvider A data provider for base map.
 */
export async function addBaseMapCountryBorder(
    mapView: MapView,
    dataProvider: BaseMapDataProvider
): Promise<void> {
    //remove country border outlines
    const theme = await mapView.getTheme();
    const tilezen = theme?.styles?.tilezen;

    if (tilezen !== undefined) {
        // Remove country border outlines on China environment
        // TODO: IP: cover with unit tests
        let outlineIndex = -1;
        tilezen.forEach((style, index) => {
            if (style.description === "outline") {
                outlineIndex = index;
            }
        });
        if (outlineIndex >= 0) {
            tilezen.splice(outlineIndex, 1);
        }

        // change the font size for sensitive Tibet places,
        // because by default ( from theme ) they are invisible till 13 zoom lvl
        if (theme.definitions !== undefined) {
            for (const key in theme.definitions) {
                if (key === "PointOfInterest.FontStyle.Size") {
                    theme.definitions[key] = 12;
                }
            }
        }
        mapView.setTheme(theme);
    }

    const countryBorderDS = new VectorTileDataSource({
        name: "DataInspector base map country borders",
        dataProvider,
        getBearerToken: dataProvider.settings.token,
        styleSetName: "here_olp_cb",
        storageLevelOffset: -2,
        maxDataLevel: 14,
        //TODO IP: hack to fix vertical line artifacts on OMV
        // https://saeljira.it.here.com/browse/HARP-12145
        roundUpCoordinatesIfNeeded: true,
    });
    await mapView.addDataSource(countryBorderDS);
    countryBorderDS.setTheme({
        styles: [
            {
                styleSet: countryBorderDS.styleSetName,
                when: [
                    "all",
                    ["==", ["get", "$geometryType"], "line"],
                    ["==", ["get", "$layer"], "boundaries"],
                    ["==", ["get", "kind"], "country"]
                ],
                technique: "label-rejection-line",
                description: "rejecting stuff",
                final: true
            }, {
                styleSet: countryBorderDS.styleSetName,
                when: [
                    "all",
                    ["==", ["get", "$geometryType"], "line"],
                    ["==", ["get", "$layer"], "boundaries"],
                    ["==", ["get", "kind"], "disputed"]
                ],
                technique: "label-rejection-line",
                description: "rejecting based on disputed border",
                final: true
            },
            // Uncomment code below for debugging purpose and change final to false
            // {
            //     styleSet: countryBorderDS.styleSetName,
            //     when: [
            //         "all",
            //         ["==", ["get", "$geometryType"], "line"],
            //         ["==", ["get", "$layer"], "boundaries"],
            //         ["==", ["get", "kind"], "country"]
            //     ],
            //     technique: "line",
            //     description:
            //         "attachment: outline, category: border-outline, render order: 880, ",
            //     renderOrder: 10000,
            //     attr: {
            //         lineWidth: 1,
            //         opacity: 1,
            //         color: "#ff0000"
            //     }
            // }
        ]
    });
}

/**
 * Sets up tile selection and hover effects for `MapView`.
 *
 * @param mapView `MapView` instance.
 * @param tileControlHandler `TileControlHandler` instance.
 */
export function setupTileSelection(mapView: MapView, tileControlHandler: TileControlHandler) {
    MapControlHandler.getInstance(mapView).addEventHandler(
        "select",
        (geoPoint, modifier) => tileControlHandler.tileSelectionHandler(geoPoint, modifier.ctrlKey)
    ).addEventHandler("zoom", (geoPoint, keyModifier, event) => {
        if (!(event instanceof MouseEvent)) {
            return;
        }
        tileControlHandler.oneLevelZoomHandler("In", event);
    }).addEventHandler("rmbdblclick", (geoPoint, keyModifier, event) => {
        if (!(event instanceof MouseEvent)) {
            return;
        }
        tileControlHandler.oneLevelZoomHandler("Out", event);
    });

    mapView.canvas.addEventListener(
        "mousemove",
        event => tileControlHandler.tileHoverHandler(event)
    );
}

/**
 * Creates authentication form and returns an instance of `AccessTokenResolver` to retrieve
 * authentication token.
 *
 * @param mapView MapView to connect.
 */
export function createTokenResolver(mapView: MapView): AccessTokenResolver {
    const authForm = new UserAuthForm({
        tokenForm: false,
        accessKeyForm: true,
        localAuthForm: true
    });

    const tokenResolver = new AccessTokenResolver(authForm, lookupServiceEnvironment);

    const container = document.getElementById("inner-container");
    if (container === null) {
        throw new Error("No container for Control Panel found.");
    }

    const controlPanel = new ControlPanel({}, container);
    controlPanel.addSection({
        widget: authForm.element
    }, 0);
    controlPanel.show();

    addConnectionListener(authForm, mapView);

    return tokenResolver;
}


/** An event handler for a click on a switch widget. */
export type FilterSwitchHandler = (
    state: boolean,
    featureType?: string,
    name?: string | number,
    attribute?: string
) => void;

/** Interface for the feature group filter. */
export interface FeatureGroupFilter {
    /** Filter switch handler. */
    switchHandler?: FilterSwitchHandler;
    /** `true` if filter is active, `false` otherwise. */
    state: boolean;
    /** `true` if group is collapsed, `false` if expanded. */
    collapsed: boolean;
    /** Feature type. */
    featureType: string;
    /** Feature attribute value. */
    name: string | number;
    /** Human-readable filter group title. */
    title: string;
    /** Array of feature attribute filters. */
    filters?: FeatureAttributeFilter[];
}

/** Interface for the feature attribute filter. */
export interface FeatureAttributeFilter {
    /** Filter switch handler. */
    switchHandler: FilterSwitchHandler;
    /** `true` if filter is active, `false` otherwise. */
    state: boolean;
    /** Feature type. */
    featureType: string;
    /** Feature attribute value. */
    name: string | number;
    /** Human-readable filter title. */
    title: string;
}

/**
 * Interface for the configuration of the [[InteractiveFilteringControl]].
 */
export interface FilteringControlProps {
    /** Id of an info panel HTML element. */
    elementId: string;

    /** CSS class for an info panel element. */
    elementClass?: string;
}

/**
 * Interface for the state configuration for the [[InteractiveFilteringControl]].
 */
export interface FilteringControlState {
    /**
     * Info panel visibility.
     */
    visible?: boolean;

    /**
     * Feature groups.
     * @example <b>Example:</b>
     * ```typescript
     * [
     *   {
     *     "state": false,
     *     "collapsed": "false",
     *     "featureType": "hazardWarning",
     *     "title": "Hazard warnings"
     *   }
     * ]
     * ```
     */
    featureGroups?: FeatureGroupFilter[];
}

/**
 * The `InteractiveFilteringControl` class represents a UI widget for filtering datasource features.
 */
export class InteractiveFilteringControl
    extends InteractiveWidget<FilteringControlProps, FilteringControlState> {

    /** @inheritDoc */
    readonly element: HTMLElement;

    /** @inheritDoc */
    readonly state: FilteringControlState;

    /** Indicates whether filters are disabled or not. */
    disabled = true;

    private m_cls = "interactive-widget interactive-filter-control widget";

    /**
     * Default constructor.
     *
     * @param props Initial settings.
     * @param state Initial state.
     */
    constructor(readonly props: FilteringControlProps, state?: FilteringControlState) {
        super();

        this.element = document.createElement("div");
        this.element.setAttribute("id", this.props.elementId);
        this.state = {};
        if (state) {
            this.setState(state);
        }
    }

    /** @inheritDoc */
    setState(
        newState: FilteringControlState,
        callback?: () => void
    ): void {
        let shouldRender = false;
        this.disabled = true;

        if (newState.featureGroups) {
            for (const featureGroup of newState.featureGroups) {
                if (this.state.featureGroups) {
                    const currentFeatureGroup = this.state.featureGroups.find((fg) => {
                        return fg.featureType === featureGroup.featureType;
                    });
                    if (currentFeatureGroup) {
                        this.calculateFeatureGroupState(featureGroup, currentFeatureGroup);
                    } else {
                        this.calculateFeatureGroupState(featureGroup);
                    }
                } else {
                    this.calculateFeatureGroupState(featureGroup);
                }
            }
            shouldRender = true;
        }

        Object.assign(this.state, newState);

        if (shouldRender) {
            this.render();
        }

        if (newState.visible !== undefined) {
            this.element.classList[(newState.visible) ? "remove" : "add"]("hidden");
        }

        this.disabled = false;

        if (typeof callback === "function") {
            callback();
        }
    }

    /** @inheritDoc */
    render() {
        this.element.innerHTML = "";
        const visibleCls = (this.state.visible) ? "" : "hidden";

        const composedClass = [this.m_cls, visibleCls, this.props.elementClass]
            .filter(el => !!el)
            .join(" ");

        if (this.state.featureGroups) {
            for (const featureGroup of this.state.featureGroups) {
                const groupElement = document.createElement("div");
                const groupElementChildren = document.createElement("div");
                const groupCaptionElement = document.createElement("div");
                const expandCollapseElement = document.createElement("div");
                const groupHeaderElement = document.createElement("h2");

                groupHeaderElement.innerHTML = featureGroup.title;

                expandCollapseElement.setAttribute(
                    "class",
                    `expand-collapse-switch ${(featureGroup.collapsed) ? "expand" : "collapse"}`
                );
                groupElementChildren.setAttribute(
                    "class",
                    `controls-inner ${(featureGroup.collapsed) ? "collapse" : "expand"}`
                );

                expandCollapseElement.addEventListener("click", () => {
                    featureGroup.collapsed = !featureGroup.collapsed;
                    expandCollapseElement.setAttribute(
                        "class",
                        `expand-collapse-switch ${(featureGroup.collapsed) ? "expand" : "collapse"}`
                    );
                    groupElementChildren.setAttribute(
                        "class",
                        `controls-inner ${(featureGroup.collapsed) ? "collapse" : "expand"}`
                    );
                });

                groupCaptionElement.setAttribute("class", "controls-caption");
                groupCaptionElement.appendChild(expandCollapseElement);
                groupCaptionElement.appendChild(groupHeaderElement);
                groupCaptionElement.appendChild(this.onOffSwitchElementGenerator(featureGroup));

                groupElement.setAttribute("class", "controls-section");
                groupElement.appendChild(groupCaptionElement);

                if (featureGroup.filters) {
                    for (const filter of featureGroup.filters) {
                        const filterEl = document.createElement("div");
                        filterEl.setAttribute("class", "control-sub-item");
                        filterEl.innerHTML = `<span class="text">${filter.title}</span>`;
                        filterEl.appendChild(this.onOffSwitchElementGenerator(filter));

                        groupElementChildren.appendChild(filterEl);
                    }
                }

                groupElement.appendChild(groupElementChildren);
                this.element.appendChild(groupElement);
            }
        }

        this.element.setAttribute("class", composedClass);
    }

    private calculateFeatureGroupState(
        newState: FeatureGroupFilter,
        prevState?: FeatureGroupFilter
    ) {
        if (prevState) {
            if (newState.state !== prevState.state) {
                this.applyFeatureGroupState(newState, newState.state);
            }
        } else {
            this.applyFeatureGroupState(newState, newState.state);
        }
    }

    private applyFeatureGroupState(newState: FeatureGroupFilter, state: boolean) {
        if (newState.switchHandler) {
            newState.switchHandler(state, newState.featureType, newState.name);
        }
    }

    private onOffSwitchElementGenerator(
        featureAttributeFilter: FeatureAttributeFilter | FeatureGroupFilter
    ): Element {
        const onOffSwitchElement = document.createElement("div");
        const checkboxEl = document.createElement("input");
        const onOffLabel = document.createElement("label");

        let checkboxId = (
            `filter-${featureAttributeFilter.featureType}-${featureAttributeFilter.name}`
        );

        checkboxEl.setAttribute("class", "onoffswitch-checkbox");
        checkboxEl.setAttribute("type", "checkbox");
        checkboxEl.setAttribute("id", checkboxId);
        if (featureAttributeFilter.state) {
            checkboxEl.setAttribute("checked", "");
        }
        checkboxEl.addEventListener("click", () => {
            featureAttributeFilter.state = !featureAttributeFilter.state;

            if (featureAttributeFilter.switchHandler) {
                featureAttributeFilter.switchHandler(
                    featureAttributeFilter.state,
                    featureAttributeFilter.featureType,
                    featureAttributeFilter.name
                );
            }
        });

        onOffSwitchElement.appendChild(checkboxEl);

        onOffLabel.setAttribute("class", "onoffswitch-label");
        onOffLabel.setAttribute("for", checkboxId);
        onOffLabel.innerHTML = (
            `<span class="onoffswitch-inner"></span><span class="onoffswitch-switch"></span>`
        );

        onOffSwitchElement.appendChild(onOffLabel);
        onOffSwitchElement.setAttribute("class", "onoffswitch");

        return onOffSwitchElement;
    }
}

export const bearerTokenProvider = () => new Promise<string>((resolve, reject) => {
    const tokenHandler = (event: MessageEvent) => {
        if (event.data.type === "userToken") {
            resolve(event.data.token);
            window.removeEventListener("message", tokenHandler);
        }
    };

    // send request to parent window to provide user token
    parent.postMessage({ type: "tokenRequest" }, "*");

    // wait and handle response with the user token
    window.addEventListener("message", tokenHandler);
});

export const isChinaEnv = (env: string): boolean => {
    return env === LookupEnvironment.HERE_CN ||
        env === LookupEnvironment.HERE_CN_DEV;
};
