/*
 * Copyright (C) 2018-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

/** @module @here/olp-examples **//***/

import { SegmentAnchorGeometryDataProvider } from "@here/geometry-data-provider";
import { DecodedTile } from "@here/harp-datasource-protocol";
import { TileKey } from "@here/harp-geoutils";
import {
    InteractiveDataSource,
    InteractiveDataSourceParameters
} from "@here/interactive-datasource-protocol";
import { SpeedLimitsTile } from "./SpeedLimitsTile";

/**
 * Initialization parameters for a [[SpeedLimitsDataSource]].
 */
export interface SpeedLimitsDataSourceParameters extends InteractiveDataSourceParameters {
    /** An instance of [[SegmentAnchorGeometryDataProvider]] to retrieve road geometries. */
    geometryDataProvider: SegmentAnchorGeometryDataProvider;
}

// tslint:disable:max-line-length
/**
 * The `SpeedLimitsDataSource` visualizes speed limits from RIB catalog. It retrieves speed limit
 * data from RIB Navigation Attributes layer and visualizes it over roads geometry from RIB Road
 * Topology layer. Speed limits data has no geometry information, it is mapped to road segments from
 * road topology layer. Roads geometry is provided by `SegmentAnchorGeometryDataProvider`.
 *
 * @see https://developer.here.com/olp/documentation/here-map-content/topics/navigation-attributes.html
 * @see https://developer.here.com/olp/documentation/here-map-content/topics/topology-layer.html
 */
// tslint:enable:max-line-length
export class SpeedLimitsDataSource extends InteractiveDataSource<SpeedLimitsTile> {
    /** Default level with data. */
    protected static readonly DEFAULT_LEVEL_WITH_DATA = 12;

    /** Name of data source to be used in logger. */
    name = "SpeedLimitsDataSource";

    /** Instance of [[SegmentAnchorGeometryDataProvider]]. */
    private m_geometryDataProvider: SegmentAnchorGeometryDataProvider;

    /**
     * Default constructor.
     *
     * @param params Data source initialization parameters.
     */
    constructor(params: SpeedLimitsDataSourceParameters) {
        super(SpeedLimitsTile, {
            name: "SpeedLimitsDataSource",
            styleSetName: "speedlimits",
            concurrentDecoderServiceName: "speedlimits",
            levelsWithData: [SpeedLimitsDataSource.DEFAULT_LEVEL_WITH_DATA],
            ...params
        });

        this.m_geometryDataProvider = params.geometryDataProvider;
    }

    /** Getter of geometry data provider to be used by tiles to map geometries. */
    get geometryDataProvider(): SegmentAnchorGeometryDataProvider {
        return this.m_geometryDataProvider;
    }

    /** @inheritDoc */
    async connect() {
        // Connect an instance of geometry data provider.
        await this.m_geometryDataProvider.connect();

        // Connect this data source to Data Service.
        await super.connect();
    }

    /** @inheritDoc */
    tileDecoded(tileKey: TileKey, decodedTile: DecodedTile): void {
        // Creates actual geometries for tile.
        this.createGeometries(tileKey, decodedTile);

        // This must be called for the `InteractiveDataSource.EVENT_TILE_DECODED` event to be fired.
        super.tileDecoded(tileKey, decodedTile);

        this.dispatchEvent({
            type: InteractiveDataSource.EVENT_BEFORE_TILE_RENDERED,
            tileKey
        });
    }
}
