/*
 * Copyright (C) 2018-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

/** @module @here/olp-examples **//***/

import {
    Geometry,
    GeometryType,
    IndexedTechnique
} from "@here/harp-datasource-protocol";
import {
    MapEnv,
    StyleSetEvaluator
} from "@here/harp-datasource-protocol/index-decoder";
import { LineGroup } from "@here/harp-lines";
import {
    DecodedInteractiveTile,
    InteractiveTile,
    IndicesRange
} from "@here/interactive-datasource-protocol";
import * as THREE from "three";
import { SpeedLimitsDataSource } from "./SpeedLimitsDataSource";
import { SpeedLimitTheme } from "./theme/speedLimitTheme";
import { TileGeometryCreator } from "@here/harp-mapview/lib/geometry/TileGeometryCreator";
import { MapObjectAdapter } from "@here/harp-mapview/lib/MapObjectAdapter";

/**
 * Reference for a feature.
 * Feature can be in a different partition.
 */
interface Reference {
    /** The partition that contains the referenced feature. */
    partition_name: string;

    /** An identifier for the referenced feature in the format of domain:system:type:id */
    identifier: string;
}

interface OrientedSegmentReference {
    /** Reference to the segment. */
    segment_ref: Reference;

    /** The referenced segment is inverted in relation to the chain. */
    inverted: boolean;
}

/** Specifies how an attribute is assigned to a linear chain of one or more Segments. */
interface SegmentAnchor {
    /**
     * The attribute's orientation relative to the chain's first-to-last orientation.
     */
    attribute_orientation: number;

    /**
     * The segments in the anchor. Will always be >= 1.
     */
    oriented_segment_ref: OrientedSegmentReference[];
}

/** Speed limit information for the segment. */
interface SpeedLimitAttribute {
    /**
     * Indices of all segment anchors to which this attribute applies. 0-based index references to
     * segment_anchor in the main partition. Never empty.
     */
    segment_anchor_index: number[];

    /**
     * The source for how the speed limit was collected.
     */
    source: number;

    /**
     * Speed limit value in kilometers per hour. If the speed limit is unlimited, the value is 0.
     */
    value: number;
}

/** Attribution layer for Navigation Attribute content. */
interface NavigationAttributesPartition {
    /**
     * All unique segment anchors in this partition. Referenced by 0-based index.
     */
    segment_anchor: SegmentAnchor[];

    /**
     * Speed limit attribution for all applicable segments in this partition.
     */
    speed_limit: SpeedLimitAttribute[];
}

/**
 * `SpeedLimitsTile` visualizes speed limit data.
 *
 * It creates actual geometry based on speed limits data from decoder and road segments coordinates
 * from road segments geometry data provider.
 */
export class SpeedLimitsTile extends InteractiveTile {
    // Threshold values for applying different techniques to speed limit geometries.
    private static readonly MEDIUM_LIMIT_LOW_VALUE = 50;
    private static readonly MEDIUM_LIMIT_HIGH_VALUE = 80;

    private m_geometries = new Array<Geometry>();
    private m_techniques = new Array<IndexedTechnique>();

    // Techniques used to visualize roads with speed limits.
    private m_lineTechniqueLow: IndexedTechnique | undefined;
    private m_lineTechniqueMedium: IndexedTechnique | undefined;
    private m_lineTechniqueHigh: IndexedTechnique | undefined;

    // Speed limits data from HMC
    private m_speedLimitsData?: NavigationAttributesPartition;

    /** @inheritdoc */
    protected createGeometries(decodedTile: DecodedInteractiveTile) {
        super.createGeometries(decodedTile);

        this.m_speedLimitsData = decodedTile.jsonData as NavigationAttributesPartition;
        // Asynchronously create geometries and call `requestUpdate` to refresh `MapView`.
        this.createSpeedLimitsGeometries().then(() => this.dataSource.requestUpdate());
    }

    /** Retrieves techniques from theme. */
    private createSpeedLimitsMaterial(): void {
        // Creating a deep clone of theme. Theme must be cloned because `ThemeEvaluator` modifies
        // theme and it will not be uniform between different tiles.
        const theme = JSON.parse(JSON.stringify(SpeedLimitTheme));
        // Create an instance of `ThemeEvaluator` to extract certain techniques from theme.
        const evaluator = new StyleSetEvaluator({ styleSet: theme.styles.speedlimits });
        // Assign techniques to an array of techniques.
        this.m_techniques = evaluator.techniques;

        // Extract all necessary techniques from theme using evaluator.
        this.m_lineTechniqueLow = evaluator.getMatchingTechniques(
            new MapEnv({ type: "speedlimit-line-low" })
        )[0];
        this.m_lineTechniqueMedium = evaluator.getMatchingTechniques(
            new MapEnv({ type: "speedlimit-line-medium" })
        )[0];
        this.m_lineTechniqueHigh = evaluator.getMatchingTechniques(
            new MapEnv({ type: "speedlimit-line-high" })
        )[0];
    }

    /** Creates geometries and materials for tile. */
    private async createSpeedLimitsGeometries(): Promise<void> {
        this.createSpeedLimitsMaterial();
        await this.createSpeedLimitsObjects();
    }

    /** Creates speed limits geometry objects. */
    private async createSpeedLimitsObjects(): Promise<void> {
        await this.createRoadGeometries();

        const decodedTile = {
            geometries: this.m_geometries,
            techniques: this.m_techniques
        };

        this.objects.length = 0;
        TileGeometryCreator.instance.createObjects(this, decodedTile, () => void 0);

        this.objects.forEach((object, index) => {
            const geometry = this.m_geometries[index];
            if (typeof geometry.objInfos?.[0] === "object") {
                object.userData = {
                    ...object.userData,
                    ...geometry.objInfos[0],
                    // geometry kind should be prefixed with datasource name
                    kind: [`${this.dataSource.name}-tileGeometry`]
                };
            }
            MapObjectAdapter.create(object, {
                technique: object.userData.technique,
                kind: object.userData.kind
            });
        });
    }

    /**
     * Creates road geometries for all speed limit items. Picks technique based on speed limit
     * value.
     */
    private async createRoadGeometries(): Promise<void> {
        const worldPos = new THREE.Vector3();

        if (this.m_speedLimitsData === undefined) {
            return;
        }

        for (const speedLimitItem of this.m_speedLimitsData.speed_limit) {
            const technique = this.getLineTechnique(speedLimitItem);
            for (const anchorIndex of speedLimitItem.segment_anchor_index) {
                // First, get actual `SegmentAnchor` from pool.
                const anchor = this.m_speedLimitsData.segment_anchor[anchorIndex];
                // Then, retrieve line coordinates from geometry data provider.
                const line = await (this.dataSource as SpeedLimitsDataSource).geometryDataProvider
                    .getSegmentAnchorCoordinates(anchor);
                const projectedLine: number[] = [];

                // Project all geo coordinates from line to world coordinates.
                for (const geoCoordinates of line) {
                    this.mapView.projection.projectPoint(
                        { latitude: geoCoordinates[1], longitude: geoCoordinates[0] },
                        worldPos
                    ).sub(this.center);
                    projectedLine.push(worldPos.x, worldPos.y, 0);
                }

                // Create geometries from projected line.
                this.createLineGeometries(
                    speedLimitItem.value,
                    projectedLine,
                    technique._index
                );
            }
        }
    }

    /**
     * Returns technique based on a certain speed limit item.
     *
     * @param speedLimitItem Source speed limit item.
     * @returns Technique for this speed limit.
     */
    private getLineTechnique(speedLimitItem: SpeedLimitAttribute): IndexedTechnique {
        if (speedLimitItem.value < SpeedLimitsTile.MEDIUM_LIMIT_LOW_VALUE) {
            return this.m_lineTechniqueLow!;
        } else if (speedLimitItem.value < SpeedLimitsTile.MEDIUM_LIMIT_HIGH_VALUE) {
            return this.m_lineTechniqueMedium!;
        } else {
            return this.m_lineTechniqueHigh!;
        }
    }

    /**
     * Creates two types of geometries for a single speed limit item - visible and invisible lines.
     *
     * @param value Speed limit value.
     * @param line Array of coordinates for roads with provided speed limit.
     * @param techniqueIndex Index of technique to render this speed limit.
     */
    private createLineGeometries(value: number, line: number[], techniqueIndex: number): void {
        const lines = new LineGroup();
        const linesCenter = new THREE.Vector3(0, 0, 0);
        lines.add(linesCenter, line, this.projection);

        const { vertices, indices } = lines;
        const buffer = new Float32Array(vertices).buffer;
        const index = new Uint32Array(indices).buffer;

        const tooltipRanges = new Map<any, IndicesRange[]>();
        tooltipRanges.set(
            `${value} km/h`,
            [{ start: 0, end: vertices.length }]
        );

        this.m_geometries.push({
            type: GeometryType.SolidLine,
            index: {
                buffer: index,
                itemCount: 1,
                type: "uint32",
                name: "index"
            },
            interleavedVertexAttributes: [{
                type: "float",
                stride: lines.stride,
                buffer,
                attributes: lines.vertexAttributes
            }],
            groups: [{
                start: 0,
                count: indices.length,
                technique: techniqueIndex,
            }],
            objInfos: [{ tooltipRanges }]
        });
    }
}
