/*
 * Copyright (C) 2020-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

import {
    GeoJsonData,
    GeoJsonDataSource,
    GeoJsonDataSourceEvent,
    GeometryColorsID
} from "@here/olp-geojson-datasource";
import * as FileUpload from "@here/fileupload";
import { DataInspector, TileBoundingBox } from "@here/interactive-mapview-ui";
import { useToken, locale, showLanguageSelector, lookupServiceEnvironment } from "../config";
import { bearerTokenProvider } from "./common";

const hostContainer = document.getElementById("map-here") as HTMLElement;
hostContainer.classList.add("full-screen");

const dataInspector = new DataInspector(
    {
        mapView: {
            theme: "../resources/normal.reduced.night.json",
            decoder: {
                url: "../decoder.bundle.js"
            }
        },
        monacoEditorWorkersBasePath: "../monaco-editor",
        translationsBasePath: "../resources/i18n/",
        streamSaverBasePath: "../resources/mitm/",
        defaultLocale: locale || undefined,
        getBearerToken: useToken ? bearerTokenProvider : undefined,
        widgets: {
            zoomLevel: false,
            statusBar: false,
            decodedPanel: {
                enabled: false
            },
            authForm: !useToken ? {
                accessKeyForm: true,
                localAuthForm: true
            } : undefined,
            layersControls: {
                enabled: false
            }
        },
        toolbar: {
            languageSelector: showLanguageSelector
        },
        lookupServiceEnvironment,
        pageUrlParams: { enabled: true }
    },
    hostContainer,
    {},
    true,
    {
        "afterupdate": addFileUploadSection,
        "aftermount": addFileUploadSection
    }
);

let fileUploadContainer: HTMLElement | undefined;
let fileUpload: Element | undefined;
let geoJsonData: GeoJsonData | undefined;
let geoJsonDataSource: GeoJsonDataSource | undefined;

function addFileUploadSection() {
    if (dataInspector.mapView === undefined) {
        return;
    }

    geoJsonDataSource = createGeoJsonDataSource();
    dataInspector.setInteractiveDataSources([geoJsonDataSource]);

    if (dataInspector.controlPanel !== undefined) {
        // Add file upload section to Control Panel.
        dataInspector.controlPanel.addSection({
            widget: getFileUploadSection(),
            name: "File Upload"
        }, 1);
        // Create new or reuse existing file upload control.
        if (fileUpload === undefined) {
            fileUpload = FileUpload.create(geoJson => {
                geoJsonData = geoJson; // save current GeoJSON data
                geoJsonDataSource?.renderGeoJson(geoJsonData); // issue rendering of GeoJSON
            });
        } else if (geoJsonData !== undefined) {
            geoJsonDataSource?.renderGeoJson(geoJsonData); // re-render GeoJSON data.
        }
        dataInspector.controlPanel.show();
    }
}

function getFileUploadSection(): HTMLElement {
    if (fileUploadContainer === undefined) {
        fileUploadContainer = document.createElement("div");
        fileUploadContainer.id = "fileupload";
    }
    return fileUploadContainer;
}

function createGeoJsonDataSource(): GeoJsonDataSource {
    const geoJsonDataSource = new GeoJsonDataSource({ themeId: GeometryColorsID.Default });
    geoJsonDataSource.addEventListener(
        GeoJsonDataSource.EVENT_BEFORE_TILE_RENDERED,
        (event: GeoJsonDataSourceEvent) => {
            const boundingBox = event.boundingBox;
            if (dataInspector.mapView === undefined || boundingBox === undefined) {
                return;
            }

            TileBoundingBox.fitMapViewToWorldBoundingBox(dataInspector.mapView, boundingBox, dataInspector.mapControls);
        }
    );
    return geoJsonDataSource;
}
