/*
 * Copyright (C) 2017-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

import { SegmentAnchorGeometryDataProvider } from "@here/geometry-data-provider";
import {
    PlatformDataProvider,
    InteractiveIntersectionAgent
} from "@here/interactive-datasource-protocol";
import {
    DataInspector,
    DataInspectorEventNames,
    PanelDisplayMode
} from "@here/interactive-mapview-ui";
import {
    useToken,
    locale,
    showLanguageSelector,
    omvDataLayer,
    omvHostHrn,
    lookupServiceEnvironment,
    ribHrn,
    ribTopologyLayer,
    speedLimitTiles
} from "../config";
import { bearerTokenProvider } from "./common";
import { SpeedLimitsDataSource } from "./datasources/speedLimitsDataSource/SpeedLimitsDataSource";

/**
 * This example showcases dynamically creating and using a custom implementation of
 * `InteractiveDataSource` - a datasource to visualize speed limits from RIB catalog.
 *
 * Because custom `SpeedLimitsDataSource` datasource is not determined by `DataInspector` by catalog
 * HRN, it is added to an instance of `DataInspector` after it is initialized.
 */

// Create an instance of `DataInspector`.
const dataInspector = new DataInspector(
    {
        elementId: "may-host-component",
        mapView: {
            theme: "../resources/normal.reduced.night.json",
            decoder: {
                url: "../decoder.bundle.js"
            },
            defaultZoomLevel: 12,
            enableTilt: false,
        },
        monacoEditorWorkersBasePath: "../monaco-editor",
        translationsBasePath: "../resources/i18n/",
        streamSaverBasePath: "../resources/mitm/",
        defaultLocale: locale || undefined,
        baseMap: {
            hrn: omvHostHrn,
            layer: omvDataLayer
        },
        getBearerToken: useToken ? bearerTokenProvider : undefined,
        widgets: {
            zoomLevel: true,
            statusBar: true,
            searchField: true,
            decodedPanel: {
                enabled: true,
                initialState: PanelDisplayMode.Expanded,
                width: 400
            },
            statsPanel: {
                enabled: true,
                initialState: PanelDisplayMode.Collapsed
            },
            authForm: !useToken ? {
                tokenForm: false,
                accessKeyForm: true,
                localAuthForm: true
            } : undefined
        },
        toolbar: {
            enabled: true,
            fullScreenButton: true,
            languageSelector: showLanguageSelector
        },
        logMessages: true,
        lookupServiceEnvironment,
        pageUrlParams: { enabled: true }
    },
    document.getElementById("map-here") as HTMLElement,
    undefined,
    true
);

let speedLimitsDataSource: SpeedLimitsDataSource | undefined;

dataInspector.addEventListener(
    DataInspectorEventNames.DATA_INSPECTOR_IS_READY,
    async () => {
        if (dataInspector.mapView === undefined || dataInspector.interactiveDataSource !== undefined) {
            return;
        }
        //create data provider
        const dataProvider = new PlatformDataProvider({
            hrn: ribHrn,
            layer: "navigation-attributes",
            getToken: dataInspector.getConfig().getBearerToken!,
            environment: lookupServiceEnvironment
        });
        await dataProvider.connect();
        // Create an instance of `SegmentAnchorGeometryDataProvider`, necessary for `SpeedLimitsDataSource`.
        const geometryDataProvider = new SegmentAnchorGeometryDataProvider(
            {
                hrn: ribHrn,
                layer: ribTopologyLayer,
                getToken: dataInspector.getConfig().getBearerToken!,
                environment: lookupServiceEnvironment
            }
        );
        await geometryDataProvider.connect();

        const selectedTiles = speedLimitsDataSource !== undefined
            ? [...speedLimitsDataSource.selectedPartitions]
            : [speedLimitTiles[0]];
        // Create a sample speed limits datasource which uses `RoadSegmentsDataProvider`.
        speedLimitsDataSource = new SpeedLimitsDataSource({
            dataProvider,
            geometryDataProvider
        });
        speedLimitsDataSource.isSelected = true;

        // An instance of `InteractiveIntersectionAgent` is needed to detect speed limit items under mouse
        // pointer.
        const intersectionAgent = new InteractiveIntersectionAgent(dataInspector.mapView);
        speedLimitsDataSource.setUpIntersectionAgent(intersectionAgent);

        await dataInspector.setInteractiveDataSources([speedLimitsDataSource]);

        // Select speed limit tiles.
        for (const tile of selectedTiles) {
            dataInspector.tileControlHandler.selectTile(tile.toString(), true);
        }
    }
);
