/*
 * Copyright (C) 2018-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

import { CoverageDataSource } from "@here/coverage-datasource";
import {
    BaseMapDataProvider,
    InteractiveIntersectionAgent,
    LookupEnvironment
} from "@here/interactive-datasource-protocol";
import {
    AccessTokenResolver,
    ControlPanel,
    FiltersPanel,
    InfoPanel,
    installInfoWidgets,
    TileControlHandler,
    UserAuthForm
} from "@here/interactive-mapview-ui";
import { RibDataProvider, RibDataSource } from "@here/rib-datasource";
import {
    useToken,
    lookupServiceEnvironment,
    ribBuildingLayer,
    ribCartoLayer,
    ribHrn,
    ribLevelWithData,
    ribTopologyLayer
} from "../config";
import {
    bearerTokenProvider,
    addConnectionListener,
    InformationPanel,
    initializeMapView,
    setupTileSelection,
    createBaseMapDataSource,
    addBaseMapCountryBorder
} from "./common";
import { MapControls } from "@here/harp-map-controls";

const canvas = document.getElementById("mapCanvas") as HTMLCanvasElement;
const mapView = initializeMapView(canvas);

// Create info panel
//tslint:disable-next-line:max-line-length
const infoPanelMsg = `Click on a map to load RIB data from '${ribTopologyLayer}', '${ribCartoLayer}' and '${ribBuildingLayer}' layers simultaneously.`;
new InformationPanel(infoPanelMsg);

const container = document.getElementById("inner-container");
if (container === null) {
    throw new Error("No container for Control Panel found.");
}

const controlPanel = new ControlPanel({}, container);

let getBearerToken;
if (useToken) {
    getBearerToken = bearerTokenProvider;
} else {
    // Create authentication form
    const authForm = new UserAuthForm({
        tokenForm: false,
        accessKeyForm: true,
        localAuthForm: true
    });
    controlPanel.addSection({
        widget: authForm.element
    }, 0);

    const tokenResolver = new AccessTokenResolver(authForm, lookupServiceEnvironment);
    getBearerToken = tokenResolver.getBearerToken.bind(tokenResolver) as () => Promise<string>;

    addConnectionListener(authForm, mapView);
}

controlPanel.show();

// Base map
createBaseMapDataSource(getBearerToken).then(omvDataSource => {
    mapView.addDataSource(omvDataSource);
    if (lookupServiceEnvironment === LookupEnvironment.HERE_CN) {
        addBaseMapCountryBorder(mapView, omvDataSource.dataProvider() as BaseMapDataProvider);
    }
});

const dataProvider = new RibDataProvider(
    {
        hrn: ribHrn,
        layer: ribTopologyLayer,
        getToken: getBearerToken,
        environment: lookupServiceEnvironment
    },
    [ribCartoLayer, ribBuildingLayer],
);

// Tile grid
const coverageDataSource = new CoverageDataSource({
    dataProvider,
    levelsWithData: [ribLevelWithData]
});
mapView.addDataSource(coverageDataSource);

// RIB data source
const ribDataSource = new RibDataSource({
    dataProvider,
    levelsWithData: [ribLevelWithData],
    intersectionAgent: new InteractiveIntersectionAgent(mapView)
});

mapView.addDataSource(ribDataSource);

// set up mouse hover event on a tile
canvas.addEventListener("mousemove", event => {
    const geoPoint = mapView.getGeoCoordinatesAt(event.pageX, event.pageY);
    if (geoPoint === null) {
        return;
    }

    const tileKey = coverageDataSource.getTilingScheme().getTileKey(geoPoint, ribLevelWithData);
    if (tileKey !== null) {
        coverageDataSource.highlightedTile = tileKey.mortonCode();
    }
});

const mapControls = new MapControls(mapView);
mapControls.zoomLevelDeltaOnDoubleClick = 0;
mapControls.setZoomLevel(12);
setupTileSelection(
    mapView,
    new TileControlHandler(mapView, mapControls, ribDataSource, coverageDataSource)
);

// create filter widget
const filtersContainer = document.createElement("div");
filtersContainer.classList.add("host-control-panel-widget-body");

const filtersPanel = new FiltersPanel({
    dataSource: ribDataSource,
}, filtersContainer);

controlPanel.addSection({ widget: filtersPanel });

const uiContainer = document.getElementById("ui-container");
if (uiContainer === null) {
    throw new Error("Element 'ui-container' not found in document");
}

installInfoWidgets(InfoPanel, ribDataSource, uiContainer, undefined, true, 0);
