/*
 * Copyright (C) 2019-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */


async function start(env, accessKeyId, accessKeySecret, scope, hrn, layerId, partitions, latLong, zoom, baseMapHRN) {
    let getToken;
    const useToken = (new URL(window.location.href).searchParams.get("useToken") === "true");
    if (useToken) {
        // Use token provided
        getToken = () => new Promise((resolve, reject) => {
            const tokenHandler = event => {
                if (event.data.type === "userToken") {
                    resolve(event.data.token);
                    window.removeEventListener("message", tokenHandler);
                }
            };

            // send request to parent window to provide user token
            parent.postMessage({ type: "tokenRequest" }, "*");

            // wait and handle response with the user token
            window.addEventListener("message", tokenHandler);
        });
    } else {
        // Get a function, that will provide a token for an access to HERE data.
        const token = new UserAuth({
            env,
            credentials: {
                accessKeyId,
                accessKeySecret
            },
            tokenRequester: requestToken,
            scope
        }).getToken();
        getToken = () => token;
    }

    // Get layer object, that will fetch GeoJSON data.
    const settings = new OlpClientSettings({
        environment: env,
        getToken
    });
    const layer = await getLayerClient(hrn, layerId, settings);
    if (!layer) {
        throw new Error("Layer not found");
    }

    // Create Leaflet map.
    const map = L.map("map", {
        renderer: L.canvas()
    });
    map.setView(latLong, zoom);

    // snippet:leaflet-get-geojson.js
    // Get tile data and add to Leaflet.
    partitions.forEach(async partition => {
        const request = new DataRequest()
            .withPartitionId(`${partition}`);
        const response = await layer.getData(request);
        const data = await response.json();

        L.geoJSON(data)
            .bindPopup(item => item.feature.properties.tooltip)
            .addTo(map);
    });
    // end:leaflet-get-geojson.js

    // Add the HERE base map. This can be replaced with the code of your preferred base map.
    const harpLeaflet = new HarpGL.HarpGL({
        theme: "https://unpkg.com/@here/harp-map-theme@0.23.0/resources/berlin_tilezen_night_reduced.json",
        fontCatalog: "https://unpkg.com/@here/harp-map-theme@0.23.0/resources/fonts/Default_FontCatalog.json"
    }).addTo(map);
    const baseMapLayer = await getLayerClient(baseMapHRN, "omv-plus-v2", settings);
    harpLeaflet.mapView.addDataSource(new harp.OmvDataSource({
        dataProvider: {
            async getTile(tileKey) {
                const request = new DataRequest()
                    .withPartitionId(tileKey.toHereTile());
                const response = await baseMapLayer.getData(request);
                return response.arrayBuffer();
            },
            async connect() { },
            async ready() { return true; },
            async register() { },
            async unregister() { }
        },
        layerName: "omv-plus-v2",
        getBearerToken: getToken,
        styleSetName: "tilezen"
    }));
    //add HERE copyright notice
    const year = new Date().getFullYear();
    const HERECopyright = (env === "here-cn") ? "HERE, NavInfo Co., Ltd" : "HERE"
    map.attributionControl.addAttribution(`&copy; <a href="https://legal.here.com/terms">${year}` +
        ` ${HERECopyright}</a>`);
}

/**
 * Fetch the layer config and return an instance of a layer client.
 *
 * @param hrn A catalog HRN.
 * @param layerId A layer ID.
 * @param settings Client settings.
 */
async function getLayerClient(hrn, layerId, settings) {
    const hrnInstance = HRN.fromString(hrn);
    const catalogClient = new CatalogClient(hrnInstance, settings);
    const config = await catalogClient.getCatalog(new CatalogRequest());
    const layerConfig = config.layers.find(item => item.id === layerId);
    if (layerConfig === undefined) {
        return;
    }

    const layerClientParams = { catalogHrn: hrnInstance, layerId, settings };
    switch (layerConfig.layerType) {
        case "versioned":
            return new VersionedLayerClient(layerClientParams);

        case "volatile":
            return new VolatileLayerClient(layerClientParams);
    }
}
