/*
 * Copyright (C) 2019-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */


async function start(env, accessKeyId, accessKeySecret, scope, hrn, layerId, partitions, latLong, zoom, baseMapHRN) {
    let getToken;
    const useToken = (new URL(window.location.href).searchParams.get("useToken") === "true");
    if (useToken) {
        // Use token provided
        getToken = () => new Promise((resolve, reject) => {
            const tokenHandler = event => {
                if (event.data.type === "userToken") {
                    resolve(event.data.token);
                    window.removeEventListener("message", tokenHandler);
                }
            };

            // send request to parent window to provide user token
            parent.postMessage({ type: "tokenRequest" }, "*");

            // wait and handle response with the user token
            window.addEventListener("message", tokenHandler);
        });
    } else {
        // snippet:leaflet-get-token.js
        // Get a function, that will provide a token to access HERE data.
        const token = new UserAuth({
            env,
            credentials: {
                accessKeyId,
                accessKeySecret
            },
            tokenRequester: requestToken,
            scope
        }).getToken();
        getToken = () => token;
        // end:leaflet-get-token.js
    }

    // snippet:leaflet-get-layer.js
    // Get layer object, that will fetch GeoJSON data.
    const settings = new OlpClientSettings({
        environment: env,
        getToken
    });
    const layerConfig = await getLayerConfig(hrn, layerId, settings);
    if (!layerConfig) {
        throw new Error("Layer config was not found");
    }

    const layer = getLayerClient(layerConfig.layerType, hrn, layerId, settings);
    // end:leaflet-get-layer.js

    // snippet:leaflet-get-decoder.js
    if (layerConfig.schema === undefined || layerConfig.schema.hrn === undefined) {
        throw new Error("Layer schema HRN is not defined");
    }

    // Get a decoder from the schema.
    const decoder = await getDecoder(layerConfig.schema.hrn, settings);
    if (!decoder) {
        throw new Error("Decoder was not found.");
    }
    // end:leaflet-get-decoder.js

    // Create Leaflet map.
    const map = L.map("map", {
        renderer: L.canvas()
    });
    map.setView(latLong, zoom);

    // snippet:leaflet-get-protobuf.js
    // Fetch, decode and visualize specific partitions.
    partitions.forEach(async partition => {
        const request = new DataRequest()
            .withPartitionId(`${partition}`);
        const response = await layer.getData(request);
        const data = await response.arrayBuffer();

        const decodedData = decode(data, decoder);

        const geojson = decodedData.segment.map(segment => ({
            type: "Feature",
            properties: {
                tooltip: segment.identifier
            },
            geometry: {
                type: "LineString",
                coordinates: segment.geometry.point.map(point => [point.longitude, point.latitude])
            }
        }));

        L.geoJSON(geojson)
            .bindPopup(layer => layer.feature.properties.tooltip)
            .addTo(map);
    });
    // end:leaflet-get-protobuf.js


    // snippet:leaflet-add-base-map.js
    // Add the HERE base map. This can be replaced with the code of your preferred base map.
    const harpLeaflet = new HarpGL.HarpGL({
        theme: "https://unpkg.com/@here/harp-map-theme@0.23.0/resources/berlin_tilezen_night_reduced.json",
        fontCatalog: "https://unpkg.com/@here/harp-map-theme@0.23.0/resources/fonts/Default_FontCatalog.json"
    }).addTo(map);
    const baseMapLayerConfig = await getLayerConfig(baseMapHRN, "omv-plus-v2", settings);
    if (!baseMapLayerConfig) {
        throw new Error("Base map layer was not found");
    }
    const baseMapLayer = getLayerClient(baseMapLayerConfig.layerType, baseMapHRN, "omv-plus-v2", settings);
    harpLeaflet.mapView.addDataSource(new harp.OmvDataSource({
        dataProvider: {
            async getTile(tileKey) {
                const request = new DataRequest()
                    .withPartitionId(tileKey.toHereTile());
                const response = await baseMapLayer.getData(request);
                return response.arrayBuffer();
            },
            async connect() { },
            async ready() { return true; },
            async register() { },
            async unregister() { },
        },
        layerName: "omv-plus-v2",
        getBearerToken: getToken,
        styleSetName: "tilezen"
    }));

    // add HERE copyright notice
    const year = new Date().getFullYear();
    const HERECopyright = (env === "here-cn") ? "HERE, NavInfo Co., Ltd" : "HERE"
    map.attributionControl.addAttribution(`&copy; <a href="https://legal.here.com/terms">${year}` +
        ` ${HERECopyright}</a>`);
    // end:leaflet-add-base-map.js
}

// snippet:leaflet-get-config-and-client.js
/**
 * Fetch the layer config and return an instance of a layer client.
 *
 * @param hrn A catalog HRN.
 * @param layerId A layer ID.
 * @param settings Client settings.
 */
async function getLayerConfig(hrn, layerId, settings) {
    const catalogClient = new CatalogClient(HRN.fromString(hrn), settings);
    const config = await catalogClient.getCatalog(new CatalogRequest());
    return config.layers.find(item => item.id === layerId);
}

/**
 * Fetch the layer config and return an instance of a layer client.
 *
 * @param layerType A layer type.
 * @param hrn A catalog HRN.
 * @param layerId A layer ID.
 * @param settings Client settings.
 */
function getLayerClient(layerType, hrn, layerId, settings) {
    const hrnInstance = HRN.fromString(hrn);
    const layerClientParams = { catalogHrn: hrnInstance, layerId, settings };
    switch (layerType) {
        case "versioned":
            return new VersionedLayerClient(layerClientParams);

        case "volatile":
            return new VolatileLayerClient(layerClientParams);

        default:
            throw new Error(`Layer type "${layerType}" is not supported yet.`);
    }
}
// end:leaflet-get-config-and-client.js

// snippet:leaflet-get-decoder-function.js
/**
 * Fetch and process schema archive to prepare a Protobuf decoder.
 *
 * @param layer Instance of a layer client.
 */
async function getDecoder(hrn, settings) {
    // Get schema with protobuf files
    const artifactClient = new ArtifactClient(settings);
    const detailsRequest = new SchemaDetailsRequest()
        .withSchema(HRN.fromString(hrn));
    const details = await artifactClient.getSchemaDetails(detailsRequest);

    if (details === undefined || details.variants === undefined) {
        return;
    }

    const variant = details.variants.find(item => item.id === "ds");
    if (variant === undefined) {
        return;
    }

    const request = new SchemaRequest().withVariant(variant);
    const archive = await artifactClient.getSchema(request);

    // Load schema as a ZIP archive
    const zip = new JSZip();
    await zip.loadAsync(archive);

    // Read all .proto file and parse them by Protobuf
    const protobufRoot = new protobuf.Root();
    Object.keys(zip.files).forEach(async fileName => {
        if (!fileName.endsWith(".proto")) {
            return;
        }

        const file = await zip.file(fileName).async("text");
        protobuf.parse(file, protobufRoot, { keepCase: true })
    });

    // Extract the manifest data.
    const manifestFile = await zip.file("META-INF/layer.manifest.json").async("text");
    const manifest = JSON.parse(manifestFile);

    return protobufRoot.lookupType(manifest.main.message);
}
// end:leaflet-get-decoder-function.js


/**
 * Decode data.
 *
 * @param data Array buffer of encoded data.
 * @param decoder Decoder.
 */
function decode(data, decoder) {
    const uint8Array = new Uint8Array(data);
    const decodedMessage = decoder.decode(uint8Array);

    return decodedMessage.$type.toObject(decodedMessage, {
        defaults: true,
        longs: String,
        enums: String,
        bytes: String,
        json: true
    });
}
