/*
* Copyright (C) 2021-2024 HERE Global B.V. and its affiliate(s). All rights reserved.
*
* This software and other materials contain proprietary information controlled by HERE and are
* protected by applicable copyright legislation. Any use and utilization of this software and other
* materials and disclosure to any third parties is conditional upon having a separate agreement
* with HERE for the access, use, utilization or disclosure of this software. In the absence of such
* agreement, the use of the software is not allowed.
*/

// Prepare styling for base map
const styleBaseMapVector = function (feature, resolution) {
    switch (feature.get("layer")) {
        case "water":
            return new ol.style.Style({
                fill: new ol.style.Fill({
                    color: "#9db9e8"
                }),
            });
        case "boundaries":
            return new ol.style.Style({
                stroke: new ol.style.Stroke({
                    color: (feature.get("kind") == "region") ? "#d0d0d0" : "#909090",
                    width: (feature.get("kind") == "region") ? 1 : 1.5
                })
            });
        case "roads":
            return new ol.style.Style({
                stroke: new ol.style.Stroke({
                    color: "#cccc55",
                    width: (feature.get("kind") == "highway") ? 2 : 1
                })
            });
        case "places":
            return new ol.style.Style({
                image: new ol.style.Circle({
                    radius: (feature.get("kind_detail") == "city") ? 2 : 0,
                    fill: new ol.style.Fill({
                        color: (feature.get("kind_detail") == "city") ? "#906060" : "#909090"
                    }),
                    stroke: 0
                }),
                text: new ol.style.Text({
                    text: (feature.get("kind_detail") == "city") ? feature.get("name") : null,
                    fill: new ol.style.Fill({
                        color: "#909090"
                    }),
                    font: "12px sans-serif"
                })
            });
        default:
            return new ol.style.Style({
                stroke: new ol.style.Stroke({
                    color: "#d0f0d0"
                }),
                fill: new ol.style.Fill({
                    color: "#e0f0e0"
                })
            });
    }
};

// Fetch the layer config and return an instance of a layer client.
async function getLayerClient(hrn, layerId, settings) {
    const hrnInstance = HRN.fromString(hrn);
    const catalogClient = new CatalogClient(hrnInstance, settings);
    const config = await catalogClient.getCatalog(new CatalogRequest());
    const layerConfig = config.layers.find(item => item.id === layerId);
    if (layerConfig === undefined) {
        return;
    }

    const layerClientParams = { catalogHrn: hrnInstance, layerId, settings };
    switch (layerConfig.layerType) {
        case "versioned":
            return new VersionedLayerClient(layerClientParams);

        case "volatile":
            return new VolatileLayerClient(layerClientParams);
    }
}

// Show an error when credentials are invalid
const fail = (form, status) => {
    status.innerText = "Cannot connect. Please double check your credentials and access.";
    form.key.disabled = false;
    form.secret.disabled = false;
    form.apikey.disabled = false;
    form.scope.disabled = false;
    form.action.disabled = false;
}

let map;
// Draw the base map when credentials are valid
const success = (layer) => {
    map = new ol.Map({
        target: "map",
        view: new ol.View({
            center: ol.proj.fromLonLat([8.41, 52.82]),
            zoom: 4
        })
    });
    map.addLayer(layer);
    document.body.removeChild(document.body.querySelector(".auth-form"));
}

// Get layer object, that will fetch GeoJSON data. Then get data and draw on map
async function getAndDrawData(env, form) {
    const hrn = "hrn:here:data::olp-here:here-geojson-samples";
    const layerId = "country-borders";
    const partitions = ["1440", "1441", "1444", "1419", "1269", "1271"];

    // Prepare authentication and token function for accessing the HERE data layer
    const authCreds = {
        env: env,
        credentials: {
            accessKeyId: form.key.value, // accessKeyId
            accessKeySecret: form.secret.value // accessKeySecret
        },
        tokenRequester: requestToken,
        scope: form.scope.value ?? "" // project scope not necessary for example data
    };
    const token = new UserAuth(authCreds).getToken();
    const getToken = () => token;

    const settings = new OlpClientSettings({
        environment: env,
        getToken
    });

    let dataLayer;
    try {
        dataLayer = await getLayerClient(hrn, layerId, settings);
        if (!dataLayer) {
            throw new Error("Layer not found");
        }
        const url = `https://vector.hereapi.com/v2/vectortiles/base/mc/{z}/{x}/{y}/omv?apikey=${form.apikey.value}`;
        const layer = new ol.layer.VectorTile({
            source: new ol.source.VectorTile({
                attributions: [
                    '<a href="https://www.here.com/legal" target="_blank">&copy; HERE</a>'
                ],
                format: new ol.format.MVT(),
                url,
                maxZoom: 17
            }),
            style: styleBaseMapVector
        });
        try {
            // check if user`s HERE API Key is valid
            const response = await fetch(url);
            if (response.status === 401) {
                fail(form, status);
            } else {
                success(layer);
            }
        } catch (e) {
            // in case fetch request is blocked by CORS policy
            success(layer);
        }
    } catch (error) {
        fail(form, status);
        throw new Error(error);
    }

    if (map === undefined) {
        return;
    }

    //var vectorSource, vectorLayer;
    partitions.forEach(async partition => {
        const request = new DataRequest()
            .withPartitionId(`${partition}`);
        const response = await dataLayer.getData(request);
        const data = await response.json();

        // Draw on map
        const vectorSource = new ol.source.Vector({
            features: new ol.format.GeoJSON().readFeatures(data, {featureProjection: "EPSG:3857"})
        });

        const vectorLayer = new ol.layer.Vector({
            source: vectorSource
        });

        map.addLayer(vectorLayer);
    });
}
