/*
 * Copyright (C) 2021 HERE Global B.V. and its affiliate(s). All rights reserved.
 *
 * This software and other materials contain proprietary information controlled by HERE and are
 * protected by applicable copyright legislation. Any use and utilization of this software and other
 * materials and disclosure to any third parties is conditional upon having a separate agreement
 * with HERE for the access, use, utilization or disclosure of this software. In the absence of such
 * agreement, the use of the software is not allowed.
 */

/**
 * Renderer plugin for LiveWeather data.
 */
(() => {
    "use strict";

    const DATA_NOT_AVAILABLE = -99999;

    const RADIANS_IN_DEGREE = 0.0174533;

    /** Arrow polygon template in arbitrary units. */
    const ARROW_SHAPE = [
        [0, 1],
        [-0.5, -1],
        [0, -0.5],
        [0, -0.5],
        [0.5, -1],
        [0, 1]
    ];

    /** Scale of the arrows. */
    const ARROW_SCALE = 0.009;

    /**
     * An offset for the visibility square. If set to 0, then the size of visibility square will be
     * equal to the tile bounding box.
     */
    const VISIBILITY_OFFSET = 0.0075;

    /**
     * An offset for the precipitation square. If set to 0, then the size of precipitation square
     * will be equal to the tile bounding box.
     */
    const PRECIPITATION_OFFSET = 0.0055;

    /**
     * An offset for the temperature square. If set to 0, then the size of temperature square will
     * be equal to the tile bounding box.
     */
    const TEMPERATURE_OFFSET = 0;

    /** Color configuration object. */
    const colors = {
        // dark grey
        unknown: [55,55,55],
        temp: [
            // purple
            {
                red: 155,
                green: 0,
                blue: 256
            },
            // blue
            {
                red: 0,
                green: 0,
                blue: 256
            },
            // white
            {
                red: 256,
                green: 256,
                blue: 256
            },
            // yellow
            {
                red: 256,
                green: 256,
                blue: 0
            },
            //red
            {
                red: 256,
                blue: 0,
                green: 0
            }
        ],
        wind: [
            // blue
            {
                red: 0,
                green: 100,
                blue: 255
            },
            // yellow
            {
                red: 255,
                green: 255,
                blue: 0
            },
            // red
            {
                red: 255,
                green: 0,
                blue: 0
            }
        ],
        precip: {
            RAIN: ["#93c701", "#ffd701", "#f05514", "#dc250e"],
            SNOW: ["#9ec8f2", "#2a8fdb", "#144bed", "#020096"],
            ICE: ["#e6a5c8", "#d24fa0", "#b71691", "#7a1570"],
            PRECIPITATION_TYPE_UNKNOWN: ["#c196e6", "#ae6ee6", "#8a32d7", "#6500ba"]
        },
        vis: "white"
    };

    /** @type {RendererPlugin} */
    const plugin = {
        /**
         * Generates GeoJSON from decoded LiveWeather data.
         *
         * @param {object} params.logger Logger to log information to console.
         * @param {object} params.layer All properties as well as dataUrl, name, partitioning and
         * originalConfiguration.
         * @param {object} params.decodeResult Input LiveWeather data decoded.
         * @param {object} params.mapUtils Various converters like `mercatorConverter`,
         * `heretileToQuad`, `hdmapCoordinateToLonLat`, etc.
         * @param {object} params.geometryDataProviders Various geometry data providers to map data.
         * @returns {object} An object with generated GeoJSON data.
         */
        getGeoJSON: function (params) {

            // Destructuring of `params` object.
            const {logger, decodeResult, mapUtils} = params;

            // Partition data decoded with the schema protobuf
            const data = decodeResult.data;
            const type = "application/vnd.geo+json; charset=utf-8";

            if (!data.weather_condition_tile) {
                logger.warn(
                    "Data is not com.here.pb.weather.v0.WeatherConditionPartition");
                return null;
            }

            const result = {
                "type": "FeatureCollection",
                "features": []
            };

            addFeatures(logger, data, mapUtils, result);
            return {
                contentType: type,
                body: result
            };
        }
    };

    return plugin;

    /**
     * Adds features for different weather attributes to the result object.
     * @param logger Logger to log information to console.
     * @param data LiveWeather data.
     * @param mapUtils An object with various util methods.
     * @param result An object to store the created features.
     */
    function addFeatures(logger, data, mapUtils, result) {
        if (!data.weather_condition_tile.length) return;

        for (let i = 0; i < data.weather_condition_tile.length; i++) {
            const wct = data.weather_condition_tile[i];
            const tileBoundingBox = heretileToBounds(wct.tile_id, mapUtils);

            addWindFeature(wct, `weather_condition_tile[${i}]`, tileBoundingBox, result);
            addVisibilityFeature(wct, `weather_condition_tile[${i}].visibility`, tileBoundingBox, result);
            addPrecipitationFeature(wct, `weather_condition_tile[${i}]`, tileBoundingBox, result);
            addTemperatureFeature(wct, `weather_condition_tile[${i}]`, tileBoundingBox, result);
        }
    }

    /**
     * Adds wind feature with directional arrow to the result object.
     * @param wct Weather condition tile.
     * @param protobufRef Protobuf ref.
     * @param tileBoundingBox Tile bounding box.
     * @param result An object to store the created feature.
     */
    function addWindFeature(wct, protobufRef, tileBoundingBox, result) {
        const wv = wct.wind_velocity;
        const wind = wv.value;

        if (wind === DATA_NOT_AVAILABLE) {
            return;
        }

        const windColor = colorGradient(
            Math.min(60, wind)/60,
            colors.wind[0],
            colors.wind[1],
            colors.wind[2]
        );
        const windOpacity = 0.2 + 0.8 * Math.min(60, wind)/60;

        const coordinates = makeArrow(
            (tileBoundingBox.topLeftLng + tileBoundingBox.bottomRightLng)/2,
            (tileBoundingBox.bottomRightLat + tileBoundingBox.topLeftLat) /2,
            wv.direction
        );

        const windFeature = {
            type: "Feature",
            properties: {
                protobufRef,
                tooltip: "Velocity: " + wind + " km/hr <br>Direction: " + wv.direction + " degrees",
                style: {
                    color: windColor,
                    opacity: windOpacity,
                    width: "1px"
                },
                featureTags: ["Wind"],
                "wind_velocity.value": wind,
                "wind_velocity.direction": wv.direction,
                "wind_velocity.confidence": wv.confidence,
                "wind_velocity.confidence_category_type": wv.confidence_category_type
            },
            geometry: {
                type: "Polygon",
                coordinates
            }
        };
        result.features.push(windFeature);
    }

    /**
     * Adds visibility feature to the result object.
     * @param wct Weather condition tile.
     * @param protobufRef Protobuf ref.
     * @param tileBoundingBox Tile bounding box.
     * @param result An object to store the created feature.
     */
    function addVisibilityFeature(wct, protobufRef, tileBoundingBox, result) {
        const vis = wct.visibility.value;

        if (vis === DATA_NOT_AVAILABLE) {
            return;
        }

        const visColor = colors.vis;
        let visOpacity;
        // Opacity for features with 600 meters visibility
        const N = 0.2;

        switch (true) {
            // visibility in range from 0 to 75 meters (thick fog)
            case vis <= 0.075:
                visOpacity = 1;
                break;
            // visibility in range from 75 to 600 meters (thin fog)
            case vis <= 0.6:
                visOpacity = (vis - 0.075) * (N - 1) / (0.6 - 0.075) + 1;
                break;
            // visibility in range from 600 to 1000 meters (almost no fog)
            case vis <= 1:
                visOpacity = -N * (vis - 0.6) / (1 - 0.6) + N;
                break;
            // visibility is greater than 1000 meters (no fog)
            default: visOpacity = 0;
        }

        const visibilityFeature = {
            type: "Feature",
            properties: {
                protobufRef,
                tooltip: `Visibility: ${vis} km`,
                style: {
                    color: visColor,
                    opacity: visOpacity,
                    width: 0
                },
                featureTags: ["Visibility"],
                "visibility.value": vis,
                "visibility.confidence": wct.visibility.confidence,
                "visibility.confidence_category_type": wct.visibility.confidence_category_type

            },
            geometry: makeSquare(tileBoundingBox, VISIBILITY_OFFSET)
        };
        result.features.push(visibilityFeature);
    }

    /**
     * Adds precipitation feature to the result object.
     * See https://en.wikipedia.org/wiki/Rain#Intensity
     *
     * @param wct Weather condition tile.
     * @param protobufRef Protobuf ref.
     * @param tileBoundingBox Tile bounding box.
     * @param result An object to store the created feature.
     */
    function addPrecipitationFeature(wct, protobufRef, tileBoundingBox, result) {
        const precip = wct.precipitation_type.precipitation_type;
        const iop = wct.iop.value;
        if (iop === DATA_NOT_AVAILABLE || precip === "NONE") {
            return;
        }

        const precipOpacity = 0.3 + 0.7 * Math.min(1, iop);
        const precipColors = colors.precip[precip];

        // Calculates index of base feature color. Possible values 0, 1, 2.
        const baseColorIndex = Math.min(2, Math.trunc(iop / 0.33));

        // Calculates the feature color between 2 default colors.
        const precipColor = colorGradient(
          Math.min(1, (iop - 0.33 * baseColorIndex) / 0.33),
          convertToRgb(precipColors[baseColorIndex]),
          convertToRgb(precipColors[baseColorIndex + 1])
        );

        const precipitationFeature = {
            type: "Feature",
            properties: {
                protobufRef,
                tooltip: precip +": "+ iop + " cm/hr",
                style: {
                    color: precipColor,
                    opacity: precipOpacity,
                    width: "0px"
                },
                featureTags: ["Precipitation"],
                "iop.value": iop,
                "iop.confidence": wct.iop.confidence,
                "iop.confidence_category_type": wct.iop.confidence_category_type,
                "precipitation_type.precipitation_type": wct.precipitation_type.precipitation_type,
                "precipitation_type.confidence": wct.precipitation_type.confidence,
                "precipitation_type.confidence_category_type": wct.precipitation_type.confidence_category_type
            },
            geometry: makeSquare(tileBoundingBox, PRECIPITATION_OFFSET)
        };
        result.features.push(precipitationFeature);
    }

    /**
     * Adds temperature feature to the result object.
     * @param wct Weather condition tile.
     * @param protobufRef Protobuf ref.
     * @param tileBoundingBox Tile bounding box.
     * @param result An object to store the created feature.
     */
    function addTemperatureFeature(wct, protobufRef, tileBoundingBox, result) {
        const { air_temperature, dew_point_temperature, tile_id } = wct;
        const temp = air_temperature.value;

        if (temp === DATA_NOT_AVAILABLE) {
            return;
        }

        const tempOpacity = 0.25;
        const tempColor = temp < 0 ?
            colorGradient(Math.min(25, Math.abs(temp))/25, colors.temp[2], colors.temp[1], colors.temp[0]) :
            colorGradient(Math.min(35, temp)/35, colors.temp[2], colors.temp[3], colors.temp[4]);

        const temperatureFeature = {
            type: "Feature",
            properties: {
                protobufRef,
                tooltip: "Air temperature: " + air_temperature.value + " °C",
                style: {
                    color: tempColor,
                    opacity: tempOpacity,
                    width: "0px"
                },
                weather_tile: tile_id,
                "air_temperature.value": air_temperature.value,
                "air_temperature.confidence": air_temperature.confidence,
                "air_temperature.confidence_category_type": air_temperature.confidence_category_type,
                "dew_point.temperature.value": dew_point_temperature.value,
                "dew_point.confidence": dew_point_temperature.confidence,
                "dew_point.confidence_category_type": dew_point_temperature.confidence_category_type,

                featureTags: ["Temperature"]
            },
            geometry: makeSquare(tileBoundingBox, TEMPERATURE_OFFSET)
        };
        result.features.push(temperatureFeature);
    }

    /**
     * Creates an arrow geometry placed in the specified location and rotated by specified amount
     * of degrees.
     *
     * @param lon Longitude.
     * @param lat Latitude.
     * @param degrees Amount of degrees to rotate the arrow clockwise. At 0 degrees, the arrow will
     * point north.
     * @return An arrow geometry.
     */
    function makeArrow(lon, lat, degrees) {
        const rotation = degrees * RADIANS_IN_DEGREE * -1;
        const position = [lon, lat];
        const ratio = getAspectRatio(position);
        const sina = Math.sin(rotation);
        const cosa = Math.cos(rotation);
        const arrowTemplateScaled = ARROW_SHAPE.map(coords => [
            coords[0] * ARROW_SCALE,
            coords[1] * ARROW_SCALE
        ]);
        const coordinates = arrowTemplateScaled.map(item => [
            (item[0] * cosa - item[1] * sina) + position[0],
            (item[0] * sina + item[1] * cosa) * ratio + position[1]
        ]);
        return [coordinates];
    }

    /**
     * Calculates an aspect ratio distortion caused by projection at certain geo coordinates.
     *
     * @param {Array} geoPoint A [longitude, latitude] coordinates pair.
     */
    function getAspectRatio(geoPoint) {
        const offset = 0.00001;
        const point = projectPoint(geoPoint);
        const pointX = projectPoint([geoPoint[0] + offset, geoPoint[1]]);
        const pointY = projectPoint([geoPoint[0], geoPoint[1] + offset]);
        const dx = Math.abs(point.x - pointX.x);
        const dy = Math.abs(point.y - pointY.y);
        return dx / dy;
    }

    /**
     * Projects a point using Mercator projection.
     *
     * @param {Array} geoPoint A [longitude, latitude] coordinates pair.
     */
    function projectPoint(geoPoint) {
        const clamped = Math.min(
            Math.max(-1.4844222297453323, geoPoint[1] * RADIANS_IN_DEGREE),
            1.4844222297453323
        );
        const projected = Math.log(Math.tan(Math.PI * 0.25 + clamped * 0.5)) / Math.PI;

        return {
            x: ((geoPoint[0] + 180) / 360) * 40075016.68557849,
            y: (projected * 0.5 + 0.5) * 40075016.68557849
        };
    }

    /**
     * Converts here tile to the bounding box.
     * @param heretile Tile ID.
     * @param mapUtils An object with various util methods.
     * @return Bounding box for the specified tile.
     */
    function heretileToBounds(heretile, mapUtils) {
        const quad = mapUtils.heretileToQuad(heretile);
        const tile = mapUtils.quadKeyToTile(quad);
        return mapUtils.uniformConverter.tileToBounds(tile.x, tile.y, tile.z);
    }

    /**
     * Creates a square for specified bounding box. The size of square is equal to the size of
     * bounding box minus offset.
     *
     * @param bb Bounding box.
     * @param offset Offset. Default value is 0.
     * @return GeoJSON polygon that represents square.
     */
    function makeSquare(bb, offset = 0) {
        return {
            type: "Polygon",
            coordinates: [[
                [bb.bottomRightLng - offset, bb.bottomRightLat - offset],
                [bb.bottomRightLng - offset, bb.topLeftLat + offset],
                [bb.topLeftLng + offset, bb.topLeftLat + offset],
                [bb.topLeftLng + offset, bb.bottomRightLat - offset],
                [bb.bottomRightLng - offset, bb.bottomRightLat - offset]
            ]]
        };
    }

    function colorGradient(fadeFraction, rgbColor1, rgbColor2, rgbColor3) {
        let color1 = rgbColor1;
        let color2 = rgbColor2;
        let fade = fadeFraction;

        // Do we have 3 colors for the gradient? Need to adjust the params.
        if (rgbColor3) {
            fade = fade * 2;

            // Find which interval to use and adjust the fade percentage
            if (fade >= 1) {
                fade -= 1;
                color1 = rgbColor2;
                color2 = rgbColor3;
            }
        }

        const diffRed = color2.red - color1.red;
        const diffGreen = color2.green - color1.green;
        const diffBlue = color2.blue - color1.blue;

        const gradient = {
            red: parseInt(Math.floor(color1.red + (diffRed * fade)), 10),
            green: parseInt(Math.floor(color1.green + (diffGreen * fade)), 10),
            blue: parseInt(Math.floor(color1.blue + (diffBlue * fade)), 10),
        };

        return 'rgb(' + gradient.red + ',' + gradient.green + ',' + gradient.blue + ')';
    }

    function convertToRgb(color) {
        return {
            red: parseInt(color[1] + color[2], 16),
            green: parseInt(color[3] + color[4], 16),
            blue: parseInt(color[5] + color[6], 16),
        };
    }

})();
